"""Utilities to help serialize Scenic objects.

The functions in this module usually do not need to be used directly.
For high-level serialization APIs, see `Scenario.sceneToBytes`,
`Scenario.simulationToBytes`, and `Scene.dumpAsScenicCode`.
"""

import io
import math
import pickle
import struct
import types

from scenic.core.distributions import Samplable, needsSampling
from scenic.core.utils import DefaultIdentityDict

## JSON


def scenicToJSON(obj):
    """Utility function to help serialize Scenic objects to JSON.

    Suitable for passing as the ``default`` argument to `json.dump`.
    At the moment this only supports very basic types like scalars and vectors:
    it does not allow encoding of an entire `Object`.
    """
    from scenic.core.vectors import Vector

    if isinstance(obj, Vector):
        return list(obj)
    raise TypeError(f"Object of type {obj.__class__.__name__} is not JSON serializable")


## Scenic code


def dumpAsScenicCode(value, stream):
    """Utility function to help export Scenic objects as Scenic code."""
    if hasattr(value, "dumpAsScenicCode"):
        value.dumpAsScenicCode(stream)
    else:
        stream.write(repr(value))


## Pickles

# If dill is installed, register some custom handlers to improve the pickling
# of Scene and Scenario objects.

try:
    import dill
except Exception:
    pass
else:
    _orig_save_module = dill.Pickler.dispatch[types.ModuleType]

    @dill.register(types.ModuleType)
    def patched_save_module(pickler, obj):
        # Save Scenic's internal modules by reference to avoid inconsistent versions
        # as well as some unpicklable objects (and shrink the size of pickles while
        # we're at it).
        name = obj.__name__
        if name == "scenic" or name.startswith("scenic."):
            pickler.save_reduce(dill._dill._import_module, (name,), obj=obj)
            return
        _orig_save_module(pickler, obj)


## Binary serialization format


class SerializationError(Exception):
    """An error occurring during serialization/deserialization of Scenic objects."""

    pass


class Serializer:
    """Class for (de)serializing scenes, etc.

    Ordinary Scenic users do not need to know about this class: they can use public
    APIs such as `Scenario.sceneToBytes`. If you have defined a custom type of
    `Distribution` whose **valueType** isn't one of the types used by the built-in
    primitive distributions (i.e. `int`, `float`, `Vector`), read on.

    The sampled value of a `Distribution` is encoded as follows:

    1. If the `Distribution` is `_deterministic`, recursively encode the sampled
       values of its dependencies.
    2. If its *valueType* is a type for which we have a "codec" (like `int` or
       `float`), use the encoding function provided by the codec.
    3. If the *valueType* provides a ``encodeTo`` method, use that.
    4. If the user has allowed the use of `pickle`, pickle the value.
    5. Otherwise raise a `SerializationError`.

    Thus, you need only extend the serialization mechanism if your `Distribution` cannot
    be made deterministic (by adding appropriate dependencies with simpler valueTypes)
    and it has an unusual **valueType**. In that case, it's best to have your **valueType**
    implement ``encodeTo`` and ``decodeFrom`` methods: see `Vector` for example. If for
    some reason you can't add those methods to the class in question, you can use
    `Serializer.addCodec` to register encoder/decoder functions. Finally, if you're only
    using serialization internally and aren't concerned about security issues or making
    the encoding as compact as possible, you can turn on the **allowPickle** option: this
    will use `pickle` to encode any objects for which no specialized encoder is known.
    """

    codecs = {}

    def __init__(self, data=b"", allowPickle=False, detectEnd=False):
        self.allowPickle = allowPickle
        self.stream = data if isinstance(data, io.BufferedIOBase) else io.BytesIO(data)
        if detectEnd:
            self.stream = io.BufferedReader(self.stream)
        self.seenObjs = set()

    def getBytes(self):
        return self.stream.getvalue()

    def atEnd(self):
        return not self.stream.peek(1)

    @classmethod
    def sceneFormatVersion(cls):
        """Current version of the `Scene` serialization format.

        Must be incremented if the `writeScene` method or any of its helper
        methods (e.g. `writeValue`) change, or if a new codec is added.
        """
        return 2

    @classmethod
    def replayFormatVersion(cls):
        """Current version of the `Simulation` replay serialization format.

        Must be incremented if the `writeReplayHeader` or `writeValue` methods
        change, or if a new codec is added.
        """
        return 2

    def writeScene(self, scenario, scene):
        """Serialize a `Scene`."""
        version = struct.pack("<H", self.sceneFormatVersion())
        self.stream.write(version)
        assert len(scenario.astHash) == 4
        self.stream.write(scenario.astHash)
        optionsHash = scenario.compileOptions.hash
        assert len(optionsHash) == 4
        self.stream.write(optionsHash)
        self.writeSample(scenario.dependencies, scene.sample)

    def readScene(self, scenario, verify=True):
        versionField = self.stream.read(2)
        if len(versionField) != 2:
            raise SerializationError("serialized Scene is corrupted")
        version = struct.unpack("<H", versionField)[0]
        if version != self.sceneFormatVersion():
            raise SerializationError(
                "cannot read serialized Scene from " "a different Scenic version"
            )
        astHash = self.stream.read(4)
        if verify and astHash != scenario.astHash:
            raise SerializationError(
                "serialized Scene does not correspond to this Scenario"
            )
        optionsHash = self.stream.read(4)
        if verify and optionsHash != scenario.compileOptions.hash:
            raise SerializationError(
                "serialized Scene used different compile options " "than this Scenario"
            )
        sample = self.readSample(scenario.dependencies)
        scene = scenario._makeSceneFromSample(sample)
        return scene

    def writeSample(self, objects, values):
        for obj in objects:
            self.writeSamplable(obj, values)

    def readSample(self, objects):
        values = DefaultIdentityDict()
        for obj in objects:
            self.readSamplable(obj, values)
        return values

    def writeSamplable(self, obj, values):
        if not needsSampling(obj):
            return  # value is not random, so no need to encode
        i = id(obj)
        if i not in self.seenObjs:
            self.seenObjs.add(i)
            obj.serializeValue(values, self)

    def readSamplable(self, obj, values):
        if not needsSampling(obj):
            return
        if obj not in values:
            values[obj] = obj.deserializeValue(self, values)

    def writeReplayHeader(self, flags):
        """Begin the encoding of a `Simulation` replay."""
        version = struct.pack("<H", self.replayFormatVersion())
        self.stream.write(version)
        self.stream.write(struct.pack("<I", flags))

    def readReplayHeader(self):
        versionField = self.stream.read(2)
        if len(versionField) != 2:
            raise SerializationError("replay is corrupted")
        version = struct.unpack("<H", versionField)[0]
        if version != self.replayFormatVersion():
            raise SerializationError("cannot read replay from a different Scenic version")
        flagsField = self.stream.read(4)
        if len(flagsField) != 4:
            raise SerializationError("replay is corrupted")
        flags = struct.unpack("<I", flagsField)[0]
        return flags

    @classmethod
    def addCodec(cls, ty, encoder, decoder):
        """Register encoder and decoder functions for the given type.

        The encoder function should have signature :samp:`encoder({value}, {stream})`
        with *stream* a :term:`binary file-like object <binary file>`. The decoder
        function should have signature :samp:`decoder({stream})` and return the decoded
        value.
        """
        if ty in cls.codecs:
            raise ValueError(f"Serializer already has a codec for type {ty}")
        cls.codecs[ty] = (encoder, decoder)

    def writeValue(self, value, ty):
        """Serialize a value of the given type."""
        try:
            if ty in self.codecs:
                encoder, decoder = self.codecs[ty]
                encoder(value, self.stream)
                return
            elif hasattr(ty, "encodeTo"):
                ty.encodeTo(value, self.stream)
                return
            elif self.allowPickle:
                pickle.dump(value, self.stream)
                return
        except Exception as e:
            raise SerializationError(
                f"failed to serialize object of type {ty.__name__}"
            ) from e

        # No known method of serialization
        raise SerializationError(f"{ty.__name__} type does not implement serialization")

    def readValue(self, ty):
        try:
            if ty in self.codecs:
                encoder, decoder = self.codecs[ty]
                return decoder(self.stream)
            elif hasattr(ty, "encodeTo"):
                return ty.decodeFrom(self.stream)
            elif self.allowPickle:
                return pickle.load(self.stream)
        except Exception as e:
            raise SerializationError(
                f"failed to deserialize object of type {ty.__name__}"
            ) from e

        # No known method of deserialization
        raise SerializationError(f"{ty.__name__} type does not implement serialization")


# Encoder/decoder functions for various types


def _writeNone(value, stream):
    pass


def _readNone(stream):
    return None


Serializer.addCodec(type(None), _writeNone, _readNone)


def writeFloat(value, stream):
    stream.write(struct.pack("<d", value))


def readFloat(stream):
    return struct.unpack("<d", stream.read(8))[0]


Serializer.addCodec(float, writeFloat, readFloat)


def writeInt(value, stream):
    # Optimize for small nonnegative integers, which commonly arise from Options
    if 0 <= value <= 252:
        stream.write(bytes([value]))
    elif -32768 <= value <= 32767:
        stream.write(bytes([253]))
        stream.write(value.to_bytes(length=2, byteorder="little", signed=True))
    elif -2147483648 <= value <= 2147483647:
        stream.write(bytes([254]))
        stream.write(value.to_bytes(length=4, byteorder="little", signed=True))
    else:
        stream.write(bytes([255]))
        length = max(1, math.ceil((value.bit_length() + 1) / 8))  # +1 for sign
        if length >= 256:
            raise SerializationError(
                "cannot serialize integers with >600 digits"
                " (what is this, cryptography?)"
            )
        stream.write(bytes([length]))
        stream.write(value.to_bytes(length=length, byteorder="little", signed=True))


def readInt(stream):
    first = stream.read(1)[0]
    if first <= 252:
        return first
    elif first == 253:
        return int.from_bytes(stream.read(2), byteorder="little", signed=True)
    elif first == 254:
        return int.from_bytes(stream.read(4), byteorder="little", signed=True)
    else:
        length = stream.read(1)[0]
        return int.from_bytes(stream.read(length), byteorder="little", signed=True)


Serializer.addCodec(int, writeInt, readInt)


def writeBool(value, stream):
    writeInt(value, stream)


def readBool(stream):
    return bool(readInt(stream))


Serializer.addCodec(bool, writeBool, readBool)


def writeBytes(value, stream):
    writeInt(len(value), stream)
    stream.write(value)


def readBytes(stream):
    length = readInt(stream)
    return stream.read(length)


Serializer.addCodec(bytes, writeBytes, readBytes)


def writeStr(value, stream):
    encoded = value.encode()
    writeBytes(encoded, stream)


def readStr(stream):
    encoded = readBytes(stream)
    return encoded.decode()


Serializer.addCodec(str, writeStr, readStr)



"""Common LGSVL interface."""

import math

import lgsvl

from scenic.core.geometry import normalizeAngle
from scenic.core.vectors import Vector


def lgsvlToScenicPosition(pos):
    """Convert LGSVL positions to Scenic positions."""
    return Vector(pos.x, pos.z)


def gpsToScenicPosition(northing, easting):
    """Convert GPS positions to Scenic positions."""
    pos = lgsvl.map_from_gps(northing, easting)
    return Vector(pos.x, pos.z)


def lgsvlToScenicElevation(pos):
    """Convert LGSVL positions to Scenic elevations."""
    return pos.y


def scenicToLGSVLPosition(pos, y=0):
    x, z = pos
    return lgsvl.Vector(x, y, z)


def lgsvlToScenicRotation(rot):
    """Convert LGSVL rotations to Scenic headings.

    Drops all but the Y component.
    """
    return normalizeAngle(-math.radians(rot.y))


def lgsvlToScenicAngularSpeed(rot):
    return -math.radians(rot.y)


def scenicToLGSVLRotation(heading):
    return lgsvl.Vector(0, -math.degrees(heading), 0)



"""Actions for dynamic agents in CARLA scenarios."""

import math as _math

import carla as _carla

from scenic.domains.driving.actions import *
import scenic.simulators.carla.utils.utils as _utils

################################################
# Actions available to all carla.Actor objects #
################################################

SetLocationAction = SetPositionAction  # TODO refactor


class SetAngularVelocityAction(Action):
    def __init__(self, angularVel):
        self.angularVel = angularVel

    def applyTo(self, obj, sim):
        xAngularVel = self.angularVel * _math.cos(obj.heading)
        yAngularVel = self.angularVel * _math.sin(obj.heading)
        newAngularVel = _utils.scalarToCarlaVector3D(xAngularVel, yAngularVel)
        obj.carlaActor.set_angular_velocity(newAngularVel)


class SetTransformAction(Action):  # TODO eliminate
    def __init__(self, pos, heading):
        self.pos = pos
        self.heading = heading

    def applyTo(self, obj, sim):
        loc = _utils.scenicToCarlaLocation(self.pos, z=obj.elevation)
        rot = _utils.scenicToCarlaRotation(self.heading)
        transform = _carla.Transform(loc, rot)
        obj.carlaActor.set_transform(transform)


#############################################
# Actions specific to carla.Vehicle objects #
#############################################


class _CarlaVehicle:
    # Mixin identifying CARLA vehicles.
    # Used to avoid importing the Vehicle class from the CARLA model, which is
    # a Scenic module not importable from Python.
    pass


class VehicleAction(Action):
    def canBeTakenBy(self, agent):
        return isinstance(agent, _CarlaVehicle)


class SetManualGearShiftAction(VehicleAction):
    def __init__(self, manualGearShift):
        if not isinstance(manualGearShift, bool):
            raise RuntimeError("Manual gear shift must be a boolean.")
        self.manualGearShift = manualGearShift

    def applyTo(self, obj, sim):
        vehicle = obj.carlaActor
        ctrl = vehicle.get_control()
        ctrl.manual_gear_shift = self.manualGearShift
        vehicle.apply_control(ctrl)


class SetGearAction(VehicleAction):
    def __init__(self, gear):
        if not isinstance(gear, int):
            raise RuntimeError("Gear must be an int.")
        self.gear = gear

    def applyTo(self, obj, sim):
        vehicle = obj.carlaActor
        ctrl = vehicle.get_control()
        ctrl.gear = self.gear
        vehicle.apply_control(ctrl)


class SetManualFirstGearShiftAction(VehicleAction):  # TODO eliminate
    def applyTo(self, obj, sim):
        ctrl = _carla.VehicleControl(manual_gear_shift=True, gear=1)
        obj.carlaActor.apply_control(ctrl)


class SetTrafficLightAction(VehicleAction):
    """Set the traffic light to desired color. It will only take
    effect if the car is within a given distance of the traffic light.

    Arguments:
        color: the string red/yellow/green/off/unknown
        distance: the maximum distance to search for traffic lights from the current position
    """

    def __init__(self, color, distance=100, group=False):
        self.color = _utils.scenicToCarlaTrafficLightStatus(color)
        if color is None:
            raise RuntimeError("Color must be red/yellow/green/off/unknown.")
        self.distance = distance

    def applyTo(self, obj, sim):
        traffic_light = obj._getClosestTrafficLight(self.distance)
        if traffic_light is not None:
            traffic_light.set_state(self.color)


class SetAutopilotAction(VehicleAction):
    def __init__(self, enabled):
        if not isinstance(enabled, bool):
            raise RuntimeError("Enabled must be a boolean.")
        self.enabled = enabled

    def applyTo(self, obj, sim):
        vehicle = obj.carlaActor
        vehicle.set_autopilot(self.enabled, sim.tm.get_port())


class SetVehicleLightStateAction(VehicleAction):
    """Set the vehicle lights' states.

    Arguments:
        vehicleLightState: Which lights are on.
    """

    def __init__(self, vehicleLightState):
        self.vehicleLightState = vehicleLightState

    def applyTo(self, obj, sim):
        obj.carlaActor.set_light_state(self.vehicleLightState)


#################################################
# Actions available to all carla.Walker objects #
#################################################


class _CarlaPedestrian:
    # Mixin identifying CARLA pedestrians. (see _CarlaVehicle)
    pass


class PedestrianAction(Action):
    def canBeTakenBy(self, agent):
        return isinstance(agent, _CarlaPedestrian)


class SetJumpAction(PedestrianAction):
    def __init__(self, jump):
        if not isinstance(jump, bool):
            raise RuntimeError("Jump must be a boolean.")
        self.jump = jump

    def applyTo(self, obj, sim):
        walker = obj.carlaActor
        ctrl = walker.get_control()
        ctrl.jump = self.jump
        walker.apply_control(ctrl)


class SetWalkAction(PedestrianAction):
    def __init__(self, enabled, maxSpeed=1.4):
        if not isinstance(enabled, bool):
            raise RuntimeError("Enabled must be a boolean.")
        self.enabled = enabled
        self.maxSpeed = maxSpeed

    def applyTo(self, obj, sim):
        controller = obj.carlaController
        if self.enabled:
            controller.start()
            controller.go_to_location(sim.world.get_random_location_from_navigation())
            controller.set_max_speed(self.maxSpeed)
        else:
            controller.stop()


class TrackWaypointsAction(Action):
    def __init__(self, waypoints, cruising_speed=10):
        self.waypoints = np.array(waypoints)
        self.curr_index = 1
        self.cruising_speed = cruising_speed

    def canBeTakenBy(self, agent):
        # return agent.lgsvlAgentType is lgsvl.AgentType.EGO
        return True

    def LQR(v_target, wheelbase, Q, R):
        A = np.matrix([[0, v_target * (5.0 / 18.0)], [0, 0]])
        B = np.matrix([[0], [(v_target / wheelbase) * (5.0 / 18.0)]])
        V = np.matrix(linalg.solve_continuous_are(A, B, Q, R))
        K = np.matrix(linalg.inv(R) * (B.T * V))
        return K

    def applyTo(self, obj, sim):
        carlaObj = obj.carlaActor
        transform = carlaObj.get_transform()
        pos = transform.location
        rot = transform.rotation
        velocity = carlaObj.get_velocity()
        th, x, y, v = (
            rot.y / 180.0 * np.pi,
            pos.x,
            pos.z,
            (velocity.x**2 + velocity.z**2) ** 0.5,
        )
        # print('state:', th, x, y, v)
        PREDICTIVE_LENGTH = 3
        MIN_SPEED = 1
        WHEEL_BASE = 3
        v = max(MIN_SPEED, v)

        x = x + PREDICTIVE_LENGTH * np.cos(-th + np.pi / 2)
        y = y + PREDICTIVE_LENGTH * np.sin(-th + np.pi / 2)
        # print('car front:', x, y)
        dists = np.linalg.norm(self.waypoints - np.array([x, y]), axis=1)
        dist_pos = np.argpartition(dists, 1)
        index = dist_pos[0]
        if index > self.curr_index and index < len(self.waypoints) - 1:
            self.curr_index = index
        p1, p2, p3 = (
            self.waypoints[self.curr_index - 1],
            self.waypoints[self.curr_index],
            self.waypoints[self.curr_index + 1],
        )

        p1_a = np.linalg.norm(p1 - np.array([x, y]))
        p3_a = np.linalg.norm(p3 - np.array([x, y]))
        p1_p2 = np.linalg.norm(p1 - p2)
        p3_p2 = np.linalg.norm(p3 - p2)

        if p1_a - p1_p2 > p3_a - p3_p2:
            p1 = p2
            p2 = p3

        # print('points:',p1, p2)
        x1, y1, x2, y2 = p1[0], p1[1], p2[0], p2[1]
        th_n = -math.atan2(y2 - y1, x2 - x1) + np.pi / 2
        d_th = (th - th_n + 3 * np.pi) % (2 * np.pi) - np.pi
        d_x = (x2 - x1) * y - (y2 - y1) * x + y2 * x1 - y1 * x2
        d_x /= np.linalg.norm(np.array([x1, y1]) - np.array([x2, y2]))
        # print('d_th, d_x:',d_th, d_x)

        K = TrackWaypoints.LQR(
            v, WHEEL_BASE, np.array([[1, 0], [0, 3]]), np.array([[10]])
        )
        u = -K * np.matrix([[-d_x], [d_th]])
        u = np.double(u)
        u_steering = min(max(u, -1), 1)

        K = 1
        u = -K * (v - self.cruising_speed)
        u_thrust = min(max(u, -1), 1)

        # print('u:', u_thrust, u_steering)

        ctrl = carlaObj.get_control()
        ctrl.steering = u_steering
        if u_thrust > 0:
            ctrl.throttle = u_thrust
        elif u_thrust < 0.1:
            ctrl.braking = -u_thrust
        carlaObj.apply_control(ctrl)



"""Assorted utility functions."""

import bz2
import collections
from contextlib import contextmanager
import functools
import itertools
import math
import os
import signal
from subprocess import CalledProcessError
import sys
import typing
import warnings
import weakref

import numpy
import trimesh

sqrt2 = math.sqrt(2)

if sys.version_info >= (3, 12):
    from itertools import batched
else:

    def batched(iterable, n):
        if n < 1:
            raise ValueError("n must be at least one")
        it = iter(iterable)
        while batch := tuple(itertools.islice(it, n)):
            yield batch


def cached(oldMethod):
    """Decorator for making a method with no arguments cache its result"""
    storageName = f"_cached_{oldMethod.__name__}"

    @functools.wraps(oldMethod)
    def wrapper(self):
        try:
            # Use __getattribute__ for direct lookup in case self is a Distribution
            return self.__getattribute__(storageName)
        except AttributeError:
            value = oldMethod(self)
            setattr(self, storageName, value)
            return value

    def clearer(self):
        try:
            delattr(self, storageName)
        except AttributeError:
            pass

    wrapper._scenic_cache_clearer = clearer

    return wrapper


_methodCaches = weakref.WeakKeyDictionary()


def cached_method(oldMethod):
    """Decorator for making a method cache its result on a per-object basis.

    Like ``functools.lru_cache(maxsize=None)`` except using a separate cache
    for each object, with the cache automatically deallocated when the object
    is garbage collected.
    """
    name = oldMethod.__name__

    @functools.wraps(oldMethod)
    def wrapper(self, *args, **kwargs):
        caches = _methodCaches.get(self, collections.defaultdict(dict))
        cachedMethod = caches.get(name)
        if cachedMethod is None:
            cachedMethod = functools.lru_cache(maxsize=None)(oldMethod)
            caches[name] = cachedMethod
        return cachedMethod(self, *args, **kwargs)

    def clearer(self):
        caches = _methodCaches.get(self, collections.defaultdict(dict))
        cachedMethod = caches.get(name)
        if cachedMethod:
            cachedMethod.cache_clear()

    wrapper._scenic_cache_clearer = clearer

    return wrapper


def cached_property(oldMethod):
    return property(cached(oldMethod))


def argsToString(args, kwargs={}):
    args = ", ".join(repr(arg) for arg in args)
    kwargs = ", ".join(f"{name}={value!r}" for name, value in kwargs.items())
    parts = []
    if args:
        parts.append(args)
    if kwargs:
        parts.append(kwargs)
    return ", ".join(parts)


@contextmanager
def alarm(seconds, handler=None, noNesting=False):
    if seconds <= 0 or not hasattr(signal, "SIGALRM"):  # SIGALRM not supported on Windows
        yield
        return
    if handler is None:
        handler = signal.SIG_IGN
    signal.signal(signal.SIGALRM, handler)
    if noNesting:
        assert oldHandler is signal.SIG_DFL, "SIGALRM handler already installed"
    length = int(math.ceil(seconds))
    previous = signal.alarm(length)
    if noNesting:
        assert previous == 0, 'nested call to "alarm"'
    try:
        yield
    finally:
        signal.alarm(0)
        signal.signal(signal.SIGALRM, signal.SIG_DFL)


def unifyMesh(mesh, verbose=False):
    """Attempt to merge mesh bodies, raising a `ValueError` if something fails.

    Should only be used with meshes that are volumes.

    If a mesh is composed of multiple bodies, the following process
    is applied:
    1. Split mesh into volumes and holes.
    2. From each volume, subtract each hole that is fully contained.
    3. Union all the resulting volumes.
    """
    assert mesh.is_volume

    # No need to unify a mesh with less than 2 bodies
    if mesh.body_count < 2:
        return mesh

    mesh_bodies = mesh.split()

    if all(m.is_volume for m in mesh_bodies):
        # If all mesh bodies are volumes, we can just return the union.
        unified_mesh = trimesh.boolean.union(mesh_bodies)

    else:
        # Split the mesh bodies into volumes and holes.
        volumes = []
        holes = []
        for m in mesh_bodies:
            if m.is_volume:
                volumes.append(m)
            else:
                m.fix_normals()
                assert m.is_volume
                holes.append(m)

        # For each volume, subtract all holes fully contained in the volume,
        # keeping track of which holes are fully contained in at least one solid.
        differenced_volumes = []
        contained_holes = set()

        for v in volumes:
            for h in filter(lambda h: h.volume < v.volume, holes):
                if h.difference(v).is_empty:
                    contained_holes.add(h)
                    v = v.difference(h)
            differenced_volumes.append(v)

        # If one or more holes was not fully contained (and thus ignored),
        # raise a warning.
        if verbose:
            if contained_holes != set(holes):
                warnings.warn(
                    "One or more holes in the provided mesh was not fully contained"
                    " in any solid (and was ignored)."
                )

        # Union all the differenced volumes together.
        unified_mesh = trimesh.boolean.union(differenced_volumes)

    # Check that the output is still a valid mesh
    if unified_mesh.is_volume:
        if verbose:
            if unified_mesh.body_count == 1:
                warnings.warn(
                    "The mesh that you loaded was composed of multiple bodies,"
                    " but Scenic was able to unify it into one single body (though"
                    " you should verify that the result is correct). To save on compile"
                    " time in the future, consider running unifyMesh on your mesh outside"
                    " of Scenic and using that output instead."
                )
            elif unified_mesh.body_count < mesh.body_count:
                warnings.warn(
                    "The mesh that you loaded was composed of multiple bodies,"
                    " but Scenic was able to unify it into fewer bodies (though"
                    " you should verify that the result is correct). To save on compile"
                    " time in the future, consider running unifyMesh on your mesh outside"
                    " of Scenic and using that output instead."
                )

        return unified_mesh
    else:
        raise ValueError("Unable to unify mesh.")


def repairMesh(mesh, pitch=(1 / 2) ** 6, verbose=True):
    """Attempt to repair a mesh, returning a proper 2-manifold.

    Repair is attempted via several steps, each sacrificing more accuracy
    but with a higher chance of returning a proper volumetric mesh.

    Repair is first attempted with easy fixes, like merging vertices and fixing
    winding. These will usually not deteriorate the quality of the mesh.

    Repair is then attempted by voxelizing the mesh, filling it, and then running
    marching cubes on the mesh. This approach is somewhat accurate but always
    produces solid objects. (This is to be expected since non watertight hollow
    objects aren't well defined).

    Repair is finally attempted by using the convex hull, which is unlikely to
    be accurate but is guaranteed to result in a volume.

    NOTE: For planar meshes, this function will throw an error.

    Args:
        mesh: The input mesh to be repaired.
        pitch: The target pitch to be used when attempting to repair the mesh via
            voxelization. A lower pitch uses smaller voxels, and thus a closer
            approximation, but can require significant additional processsing time.
            The actual pitch used may be larger if needed to get a manifold mesh.
        verbose: Whether or not to print warnings describing attempts to repair the mesh.
    """
    # If mesh is already a volume, we're done.
    if mesh.is_volume:
        return mesh

    # If mesh is planar, we can't fix it.
    if numpy.any(mesh.extents == 0):
        raise ValueError("repairMesh is undefined for planar meshes.")

    # Pitch must be positive
    if pitch <= 0:
        raise ValueError("pitch parameter must be positive.")

    ## Trimesh Processing ##
    processed_mesh = mesh.process(validate=True, merge_tex=True, merge_norm=True).copy()

    if processed_mesh.is_volume:
        if verbose:
            warnings.warn("Mesh was repaired via Trimesh process function.")

        return processed_mesh

    if verbose:
        warnings.warn(
            "Mesh could not be repaired via Trimesh process function."
            " attempting voxelization + marching cubes."
        )

    ## Voxelization + Marching Cubes ##
    # Extract largest dimension and scale so that it is unit length
    dims = numpy.abs(processed_mesh.extents)
    position = processed_mesh.bounding_box.center_mass
    scale = numpy.max(dims)

    processed_mesh.vertices -= position

    scale_matrix = numpy.eye(4)
    scale_matrix[:3, :3] /= scale
    processed_mesh.apply_transform(scale_matrix)

    # Compute new mesh with largest possible pitch
    curr_pitch = pitch

    while curr_pitch < 1:
        new_mesh = processed_mesh.voxelized(pitch=curr_pitch).fill().marching_cubes

        if new_mesh.is_volume:
            if verbose:
                warnings.warn(
                    f"Mesh was repaired via voxelization + marching cubes"
                    f" with pitch {curr_pitch}. Check the output to ensure it"
                    f" looks reasonable."
                )

            # Center new mesh
            new_mesh.vertices -= new_mesh.bounding_box.center_mass

            # Rescale mesh to original size
            orig_scale = new_mesh.extents / dims
            scale_matrix = numpy.eye(4)
            scale_matrix[:3, :3] /= orig_scale
            new_mesh.apply_transform(scale_matrix)

            # # Return mesh center to original position
            new_mesh.vertices += position

            return new_mesh

        curr_pitch *= 2

    raise ValueError("Mesh could not be repaired.")


class DefaultIdentityDict:
    """Dictionary which is the identity map by default.

    The map works on all objects, even unhashable ones, but doesn't support all
    of the standard mapping operations.
    """

    def __init__(self):
        self.storage = {}

    def clear(self):
        self.storage.clear()

    def __getitem__(self, key):
        return self.storage.get(id(key), key)

    def __setitem__(self, key, value):
        self.storage[id(key)] = value

    def __contains__(self, key):
        return id(key) in self.storage

    def __repr__(self):
        pairs = (f"{hex(key)}: {value!r}" for key, value in self.storage.items())
        allPairs = ", ".join(pairs)
        return f"<DefaultIdentityDict {{{allPairs}}}>"


# Patched version of typing.get_type_hints fixing bpo-37838

if sys.version_info >= (3, 8, 1) or (
    sys.version_info < (3, 8) and sys.version_info >= (3, 7, 6)
):
    get_type_hints = typing.get_type_hints
else:
    import types

    def get_type_hints(obj, globalns=None, localns=None):
        if not isinstance(obj, (type, types.ModuleType)) and globalns is None:
            wrapped = obj
            while hasattr(wrapped, "__wrapped__"):
                wrapped = wrapped.__wrapped__
            globalns = getattr(wrapped, "__globals__", {})
        return typing.get_type_hints(obj, globalns, localns)



"""Actions for agents in the LGSVL model."""

import math

import lgsvl
import numpy as np
from scipy import linalg

from scenic.core.type_support import toVector
from scenic.domains.driving.actions import *  # Most actions imported here
import scenic.simulators.lgsvl.utils as utils
import scenic.syntax.veneer as veneer


class FollowWaypointsAction(Action):
    def __init__(self, waypoints):
        self.waypoints = tuple(waypoints)
        if not isinstance(self.waypoints[0], lgsvl.DriveWaypoint):
            pts = []
            for wp in self.waypoints:
                elev = veneer.simulation().groundElevationAt(wp.position)
                pos = utils.scenicToLGSVLPosition(wp.position, y=elev)
                rot = utils.scenicToLGSVLRotation(wp.heading)
                pt = lgsvl.DriveWaypoint(pos, wp.speed, rot)
                pts.append(pt)
            self.waypoints = tuple(pts)

        self.lastTime = -2

    def canBeTakenBy(self, agent):
        return agent.lgsvlAgentType in (lgsvl.AgentType.NPC, lgsvl.AgentType.PEDESTRIAN)

    def applyTo(self, obj, sim):
        if sim.currentTime is not self.lastTime + 1:
            obj.lgsvlObject.follow(self.waypoints)
        self.lastTime = sim.currentTime


class CancelWaypointsAction(Action):
    def canBeTakenBy(self, agent):
        return agent.lgsvlAgentType in (lgsvl.AgentType.NPC, lgsvl.AgentType.PEDESTRIAN)

    def applyTo(self, obj, sim):
        obj.lgsvlObject.walk_randomly(False)


class SetDestinationAction(Action):
    def __init__(self, dest):
        self.dest = toVector(dest)
        self.timer = 0

    def canBeTakenBy(self, agent):
        return bool(getattr(agent, "dreamview", False))

    def applyTo(self, obj, sim):
        if self.timer == 0:
            obj.dreamview.setup_apollo(self.dest.x, self.dest.y, obj.apolloModules)

        # push vehicle for 1 second to start
        oneSec = int(1.0 / sim.timestep)
        if self.timer < oneSec:
            obj.control.throttle = 0.5
        elif self.timer == oneSec:
            obj.control.throttle = 0.5
            obj._stickyControl = False
        self.timer = self.timer + 1


class TrackWaypointsAction(Action):
    def __init__(self, waypoints, cruising_speed=10):
        self.waypoints = np.array(waypoints)
        self.curr_index = 1
        self.cruising_speed = cruising_speed

    def canBeTakenBy(self, agent):
        return agent.lgsvlAgentType is lgsvl.AgentType.EGO

    def LQR(v_target, wheelbase, Q, R):
        A = np.matrix([[0, v_target * (5.0 / 18.0)], [0, 0]])
        B = np.matrix([[0], [(v_target / wheelbase) * (5.0 / 18.0)]])
        V = np.matrix(linalg.solve_continuous_are(A, B, Q, R))
        K = np.matrix(linalg.inv(R) * (B.T * V))
        return K

    def applyTo(self, obj, sim):
        lgsvlObject = obj.lgsvlObject
        state = lgsvlObject.state
        pos = state.transform.position
        rot = state.transform.rotation
        velocity = state.velocity
        th, x, y, v = (
            rot.y / 180.0 * np.pi,
            pos.x,
            pos.z,
            (velocity.x**2 + velocity.z**2) ** 0.5,
        )
        # print('state:', th, x, y, v)
        PREDICTIVE_LENGTH = 3
        MIN_SPEED = 1
        WHEEL_BASE = 3
        v = max(MIN_SPEED, v)

        x = x + PREDICTIVE_LENGTH * np.cos(-th + np.pi / 2)
        y = y + PREDICTIVE_LENGTH * np.sin(-th + np.pi / 2)
        # print('car front:', x, y)
        dists = np.linalg.norm(self.waypoints - np.array([x, y]), axis=1)
        dist_pos = np.argpartition(dists, 1)
        index = dist_pos[0]
        if index > self.curr_index and index < len(self.waypoints) - 1:
            self.curr_index = index
        p1, p2, p3 = (
            self.waypoints[self.curr_index - 1],
            self.waypoints[self.curr_index],
            self.waypoints[self.curr_index + 1],
        )

        p1_a = np.linalg.norm(p1 - np.array([x, y]))
        p3_a = np.linalg.norm(p3 - np.array([x, y]))
        p1_p2 = np.linalg.norm(p1 - p2)
        p3_p2 = np.linalg.norm(p3 - p2)

        if p1_a - p1_p2 > p3_a - p3_p2:
            p1 = p2
            p2 = p3

        # print('points:',p1, p2)
        x1, y1, x2, y2 = p1[0], p1[1], p2[0], p2[1]
        th_n = -math.atan2(y2 - y1, x2 - x1) + np.pi / 2
        d_th = (th - th_n + 3 * np.pi) % (2 * np.pi) - np.pi
        d_x = (x2 - x1) * y - (y2 - y1) * x + y2 * x1 - y1 * x2
        d_x /= np.linalg.norm(np.array([x1, y1]) - np.array([x2, y2]))
        # print('d_th, d_x:',d_th, d_x)

        K = TrackWaypoints.LQR(
            v, WHEEL_BASE, np.array([[1, 0], [0, 3]]), np.array([[10]])
        )
        u = -K * np.matrix([[-d_x], [d_th]])
        u = np.double(u)
        u_steering = min(max(u, -1), 1)

        K = 1
        u = -K * (v - self.cruising_speed)
        u_thrust = min(max(u, -1), 1)

        # print('u:', u_thrust, u_steering)

        cntrl = lgsvl.VehicleControl()
        cntrl.steering = u_steering
        if u_thrust > 0:
            cntrl.throttle = u_thrust
        elif u_thrust < 0.1:
            cntrl.braking = -u_thrust
        lgsvlObject.apply_control(cntrl, True)



"""Pruning parts of the sample space which violate requirements.

The top-level function here, `prune`, is called as the very last step of scenario
compilation (from `translator.constructScenarioFrom`).
"""

import builtins
import collections
import math
import time

import numpy
import shapely.geometry
import shapely.geos
from trimesh.transformations import translation_matrix

from scenic.core.distributions import (
    AttributeDistribution,
    FunctionDistribution,
    MethodDistribution,
    OperatorDistribution,
    Samplable,
    dependencies,
    needsSampling,
    supportInterval,
    underlyingFunction,
)
from scenic.core.errors import InvalidScenarioError
from scenic.core.geometry import hypot, normalizeAngle, plotPolygon, polygonUnion
from scenic.core.object_types import Object, Point
import scenic.core.regions as regions
from scenic.core.regions import (
    EmptyRegion,
    MeshSurfaceRegion,
    MeshVolumeRegion,
    PolygonalRegion,
    Region,
    VoxelRegion,
)
from scenic.core.type_support import TypecheckedDistribution
from scenic.core.vectors import (
    PolygonalVectorField,
    Vector,
    VectorField,
    VectorMethodDistribution,
    VectorOperatorDistribution,
)
from scenic.core.workspaces import Workspace
from scenic.syntax.relations import DistanceRelation, RelativeHeadingRelation

### Constants
PRUNING_PITCH = 0.15


### Utilities
def currentPropValue(obj, prop):
    """Get the current value of an object's property, taking into account prior pruning."""
    value = getattr(obj, prop)
    return value._conditioned if isinstance(value, Samplable) else value


def isMethodCall(thing, method):
    """Match calls to a given method, taking into account distribution decorators."""
    if not isinstance(thing, (MethodDistribution, VectorMethodDistribution)):
        return False
    return thing.method is underlyingFunction(method)


def isFunctionCall(thing, function):
    """Match calls to a given function, taking into account distribution decorators."""
    if not isinstance(thing, FunctionDistribution):
        return False
    return thing.function is underlyingFunction(function)


def unpackWorkspace(reg):
    if isinstance(reg, Workspace):
        return reg.region
    else:
        return reg


def matchInRegion(position):
    """Match uniform samples from a `Region`

    Returns the Region, if any, the offset that should be added to the base, and
    the PointInRegionDistribution itself.
    """
    # Case 1: Position is simply a point in a region
    if isinstance(position, regions.PointInRegionDistribution):
        reg = unpackWorkspace(position.region)
        return reg, None, position

    # Case 2: Position is a point in a region with a vector offset.
    if isinstance(position, VectorOperatorDistribution) and position.operator in (
        "__add__",
        "__radd__",
    ):
        if isinstance(position.object, regions.PointInRegionDistribution):
            reg = unpackWorkspace(position.object.region)
            assert len(position.operands) == 1
            offset = position.operands[0]

            return reg, offset, position.object

    return None, None, None


def matchPolygonalField(heading, position):
    """Match orientation yaw defined by a `PolygonalVectorField` at the given position.

    Matches the yaw attribute of orientations exactly equal to a `PolygonalVectorField`,
    or offset by a bounded disturbance. Returns a triple consisting of the matched field
    if any, together with lower/upper bounds on the disturbance.
    """
    if isFunctionCall(heading, normalizeAngle):
        assert len(heading.arguments) == 1
        return matchPolygonalField(heading.arguments[0], position)

    if (
        isinstance(heading, TypecheckedDistribution)
        and heading._valueType is builtins.float
    ):
        return matchPolygonalField(heading._dist, position)

    if (
        isinstance(heading, AttributeDistribution) and heading.attribute == "yaw"
    ):  # TODO generalize to other 3D angles?
        orientation = heading.object
        if (
            isMethodCall(orientation, VectorField.__getitem__)
            and isinstance(orientation.object, PolygonalVectorField)
            and orientation.arguments == (position,)
        ):
            return orientation.object, 0, 0

    if isinstance(heading, OperatorDistribution) and heading.operator in (
        "__add__",
        "__radd__",
    ):
        field, lower, upper = matchPolygonalField(heading.object, position)
        if field is not None:
            assert len(heading.operands) == 1
            offset = heading.operands[0]
            ol, oh = supportInterval(offset)
            if ol is not None and oh is not None:
                return field, lower + ol, upper + oh

    return None, 0, 0


### Pruning procedures


def prune(scenario, verbosity=1):
    """Prune a `Scenario`, removing infeasible parts of the space.

    This function directly modifies the Distributions used in the Scenario,
    but leaves the conditional distribution under the scenario's requirements
    unchanged. See `Samplable.conditionTo`.

    Currently, the following pruning techniques are applied in order:

        * Pruning based on containment (`pruneContainment`)
        * Pruning based on relative heading bounds (`pruneRelativeHeading`)
    """
    if verbosity >= 1:
        print("  Pruning scenario...")
        startTime = time.time()

    pruneContainment(scenario, verbosity)
    pruneRelativeHeading(scenario, verbosity)
    pruneVisibility(scenario, verbosity)

    if verbosity >= 1:
        totalTime = time.time() - startTime
        print(f"  Pruned scenario in {totalTime:.4g} seconds.")


## Pruning based on containment
def pruneContainment(scenario, verbosity):
    """Prune based on the requirement that individual Objects fit within their container.

    Specifically, if O is positioned uniformly (with a possible offset) in region B and
    has container C, then we can instead pick a position uniformly in their intersection.
    If we can also lower bound the radius of O, then we can first erode C by that distance
    minus that maximum offset distance.
    """
    for obj in scenario.objects:
        # Extract the base region and container region, while doing minor checks.
        base, offset, _ = matchInRegion(obj.position)

        if base is None or needsSampling(base):
            continue

        if isinstance(base, regions.EmptyRegion):
            raise InvalidScenarioError(f"Object {obj} placed in empty region")

        container = scenario.containerOfObject(obj)

        if container is None or needsSampling(container):
            continue

        if isinstance(container, regions.EmptyRegion):
            raise InvalidScenarioError(f"Object {obj} contained in empty region")

        # Compute the maximum distance the object can be from the sampled point
        if offset is not None:
            # TODO: Support interval doesn't really work here for random values.
            if isinstance(base, PolygonalRegion):
                # Special handling for 2D regions that ignores vertical component of offset
                offset_2d = Vector(offset.x, offset.y, 0)
                _, maxDistance = supportInterval(offset_2d.norm())
            else:
                _, maxDistance = supportInterval(offset.norm())
        else:
            maxDistance = 0

        # Compute the minimum radius of the object, with respect to the
        # bounded dimensions of the container.
        if (
            isinstance(base, PolygonalRegion)
            and supportInterval(obj.pitch) == (0, 0)
            and supportInterval(obj.roll) == (0, 0)
        ):
            # Special handling for 2D regions with no pitch or roll,
            # using planar inradius instead.
            minRadius, _ = supportInterval(obj.planarInradius)
        else:
            # For most regions, use full object inradius.
            minRadius, _ = supportInterval(obj.inradius)

        # Erode the container if possible and productive
        if (
            maxDistance is not None
            and minRadius is not None
            and (maxErosion := minRadius - maxDistance) > 0
        ):
            if hasattr(container, "buffer"):
                # We can do an exact erosion
                container = container.buffer(-maxErosion)
            elif isinstance(container, MeshVolumeRegion):
                current_pitch = PRUNING_PITCH
                eroded_container = None

                while eroded_container is None:
                    # We can attempt to erode a voxel approximation of the MeshVolumeRegion.
                    eroded_container = container._erodeOverapproximate(
                        maxErosion, PRUNING_PITCH
                    )

                    if isinstance(eroded_container, VoxelRegion):
                        eroded_container = eroded_container.mesh

                    current_pitch = min(2 * current_pitch, 1)

                # Now check if this erosion is valid and useful, i.e. do we have less volume
                # to sample from. If so, replace the original container.
                if (
                    eroded_container is not None
                    and eroded_container.size < container.size
                ):
                    container = eroded_container

        # Restrict the base region to the possibly eroded container, unless
        # they're the same in which case we're done
        if base is container:
            continue

        newBase = base.intersect(container)
        newBase.orientation = base.orientation

        # Check if base was a volume and newBase is a surface,
        # in which case the mesh operation might be undefined and we abort.
        if isinstance(base, MeshVolumeRegion) and isinstance(newBase, MeshSurfaceRegion):
            continue

        # Check newBase properties
        if isinstance(newBase, EmptyRegion):
            raise InvalidScenarioError(f"Object {obj} does not fit in container")

        percentage_pruned = percentagePruned(base, newBase)

        if percentage_pruned is None:
            if verbosity >= 1:
                print(
                    f"    Region containment constraint pruning attempted but could not compute percentage for {base} and {newBase}."
                )
        else:
            if percentage_pruned <= 0.001:
                # We didn't really prune anything, don't bother setting new position
                continue

            if verbosity >= 1:
                print(
                    f"    Region containment constraint pruned {percentage_pruned:.1f}% of space."
                )

        # Condition object to pruned position
        newPos = regions.Region.uniformPointIn(newBase)

        if offset is not None:
            newPos += offset

        obj.position.conditionTo(newPos)


## Pruning based on orientation
def pruneRelativeHeading(scenario, verbosity):
    """Prune based on requirements bounding the relative heading of an Object.

    Specifically, if an object O is:

        * positioned uniformly within a polygonal region B;
        * aligned to a polygonal vector field F (up to a bounded offset);

    and another object O' is:

        * aligned to a polygonal vector field F' (up to a bounded offset);
        * at most some finite maximum distance from O;
        * required to have relative heading within a bounded offset of that of O;

    then we can instead position O uniformly in the subset of B intersecting the cells
    of F which satisfy the relative heading requirements w.r.t. some cell of F' which
    is within the distance bound.
    """
    # TODO Add test for empty pruned polygon (Might cause crash?)
    # Check which objects are (approximately) aligned to polygonal vector fields
    fields = {}
    for obj in scenario.objects:
        field, offsetL, offsetR = matchPolygonalField(obj.heading, obj.position)
        if field is not None:
            fields[obj] = (field, offsetL, offsetR)

    # Check for relative heading relations among such objects
    for obj, (field, offsetL, offsetR) in fields.items():
        position = currentPropValue(obj, "position")
        base, offset, _ = matchInRegion(position)

        # obj must be positioned uniformly in a Region
        if base is None or needsSampling(base):
            continue

        if offset is not None:
            continue

        basePoly = regions.toPolygon(base)
        if basePoly is None:  # the Region must be polygonal
            continue

        newBasePoly = basePoly
        for rel in obj._relations:
            if isinstance(rel, RelativeHeadingRelation) and rel.target in fields:
                tField, tOffsetL, tOffsetR = fields[rel.target]
                maxDist = maxDistanceBetween(scenario, obj, rel.target)
                if maxDist == float("inf"):
                    # the distance between the objects must be bounded
                    continue
                feasible = feasibleRHPolygon(
                    field,
                    offsetL,
                    offsetR,
                    tField,
                    tOffsetL,
                    tOffsetR,
                    rel.lower,
                    rel.upper,
                    maxDist,
                )
                if feasible is None:
                    # the RH bounds may be too weak to restrict the space
                    continue
                try:
                    pruned = newBasePoly & feasible
                except shapely.geos.TopologicalError:  # TODO how can we prevent these??
                    pruned = newBasePoly & feasible.buffer(0.1, cap_style=2)
                if verbosity >= 1:
                    percent = 100 * (1.0 - (pruned.area / newBasePoly.area))
                    print(
                        f"    Relative heading constraint pruned {percent:.1f}% of space."
                    )
                newBasePoly = pruned

        if newBasePoly is not basePoly:
            newBase = regions.PolygonalRegion(
                polygon=newBasePoly, orientation=base.orientation
            )
            newPos = regions.Region.uniformPointIn(newBase)
            obj.position.conditionTo(newPos)


# Pruning based on visibility
def pruneVisibility(scenario, verbosity):
    ego = scenario.egoObject

    for obj in scenario.objects:
        # Extract the base region if it exists
        position = currentPropValue(obj, "position")
        base, offset, pir_dist = matchInRegion(position)

        # Compute the maximum distance the object can be from the sampled point
        if offset is not None:
            _, maxDistance = supportInterval(offset.norm())
        else:
            maxDistance = 0

        if (
            base is None
            or needsSampling(base)
            or needsSampling(obj.radius)
            or maxDistance is None
        ):
            continue

        currBase = base
        currDist = pir_dist
        currPos = position

        # Define a helper function to attempt to buffer an observer's visibleRegion, resulting
        # in a region that contains all points that could feasibly be the position
        # of obj, if it is visible from the observer. If possible buffer exactly, otherwise
        # try to buffer approximately, and if that is also not feasible just return the viewRegion.
        def bufferHelper(viewRegion):
            buffer_quantity = obj.radius + maxDistance
            if hasattr(viewRegion, "buffer"):
                return viewRegion.buffer(buffer_quantity)
            elif hasattr(viewRegion, "_bufferOverapproximate"):
                if needsSampling(viewRegion):
                    return viewRegion._bufferOverapproximate(buffer_quantity, 1)
                else:
                    current_pitch = PRUNING_PITCH
                    buffered_container = None

                    while buffered_container is None:
                        buffered_container = viewRegion._bufferOverapproximate(
                            buffer_quantity, current_pitch
                        )

                        if isinstance(buffered_container, VoxelRegion):
                            buffered_container = buffered_container.mesh

                        current_pitch = min(2 * current_pitch, 1)

                    assert buffered_container is not None

                    return buffered_container
            else:
                assert False

        # Prune based off visibility/non-visibility requirements
        if obj.requireVisible and obj is not ego:
            # We can restrict the base region to the buffered visible region
            # of the ego.
            if verbosity >= 1:
                print(
                    f"    Pruning restricted base region of {obj} to visible region of ego."
                )
            candidateBase = currBase.intersect(bufferHelper(ego.visibleRegion))
            candidateDist = regions.Region.uniformPointIn(candidateBase)

            # Condition object to pruned position
            if offset is not None:
                candidatePos = candidateDist + offset
            else:
                candidatePos = candidateDist

            if not checkConditionedCycle(candidatePos, currPos):
                currBase = candidateBase
                currDist = candidateDist
                if offset is not None:
                    currPos = currDist + offset
                else:
                    currPos = currDist

        if obj._observingEntity:
            # We can restrict the base region to the buffered visible region
            # of the observing entity.
            if verbosity >= 1:
                print(
                    f"    Pruning restricted base region of {obj} to visible region of {obj._observingEntity}."
                )
            candidateBase = currBase.intersect(
                bufferHelper(obj._observingEntity.visibleRegion)
            )
            candidateDist = regions.Region.uniformPointIn(candidateBase)

            if not checkConditionedCycle(candidateDist, currDist):
                currBase = candidateBase
                currDist = candidateDist
                if offset is not None:
                    currPos = currDist + offset
                else:
                    currPos = currDist

        # Check currBase properties
        if isinstance(currBase, EmptyRegion):
            raise InvalidScenarioError(
                f"Object {obj} can not satisfy visibility/non-visibility constraints."
            )

        percentage_pruned = percentagePruned(base, currBase)

        if percentage_pruned is None:
            if verbosity >= 1:
                print(
                    f"    Visibility pruning attempted but could not compute percentage for {base} and {currBase}."
                )
        else:
            if percentage_pruned <= 0.001:
                # We didn't really prune anything, skip conditioning
                continue

            if verbosity >= 1:
                print(f"    Visibility pruning pruned {percentage_pruned:.1f}% of space.")

        # Condition position value to pruned position
        obj.position.conditionTo(currPos)


def maxDistanceBetween(scenario, obj, target):
    """Upper bound the distance between the given Objects."""
    # visDist is initialized to infinity. Then we can use
    # various visibility constraints to upper bound it,
    # keeping the tightest bound.
    ego = scenario.egoObject
    visDist = float("inf")

    if obj is ego and target.requireVisible:
        visDist = min(visDist, visibilityBound(ego, target))
    if target is ego and obj.requireVisible:
        visDist = min(visDist, visibilityBound(ego, obj))
    if obj._observingEntity is target:
        visDist = min(visDist, visibilityBound(target, obj))
    if target._observingEntity is obj:
        visDist = min(visDist, visibilityBound(obj, target))

    # Check for any distance bounds implied by user-specified requirements
    reqDist = float("inf")
    for rel in obj._relations:
        if isinstance(rel, DistanceRelation) and rel.target is target:
            if rel.upper < reqDist:
                reqDist = rel.upper

    return min(visDist, reqDist)


def visibilityBound(obj, target):
    """Upper bound the distance from an Object to another it can see."""
    # Upper bound on visible distance is a sum of several terms:
    # 1. obj.visibleDistance
    _, maxVisibleDistance = supportInterval(obj.visibleDistance)
    if maxVisibleDistance is None:
        return None
    # 2. distance from obj's center to its camera
    _, maxCameraX = supportInterval(obj.cameraOffset.x)
    _, maxCameraY = supportInterval(obj.cameraOffset.y)
    if maxCameraX is None or maxCameraY is None:
        return None
    maxVisibleDistance += math.hypot(maxCameraX, maxCameraY)
    # 3. radius of target
    _, maxRadius = supportInterval(target.radius)
    if maxRadius is None:
        return None
    maxVisibleDistance += maxRadius
    return maxVisibleDistance


def feasibleRHPolygon(
    field, offsetL, offsetR, tField, tOffsetL, tOffsetR, lowerBound, upperBound, maxDist
):
    """Find where objects aligned to the given fields can satisfy the given RH bounds."""
    if (
        offsetR - offsetL >= math.tau
        or tOffsetR - tOffsetL >= math.tau
        or upperBound - lowerBound >= math.tau
    ):
        return None
    polygons = []
    expanded = [(poly.buffer(maxDist), heading) for poly, heading in tField.cells]
    for baseCell, baseHeading in field.cells:
        # TODO skip cells not contained in base region?
        for expandedTargetCell, targetHeading in expanded:
            lower, upper = relativeHeadingRange(
                baseHeading, offsetL, offsetR, targetHeading, tOffsetL, tOffsetR
            )
            if upper >= lowerBound and lower <= upperBound:  # RH intervals overlap
                intersection = baseCell & expandedTargetCell
                if not intersection.is_empty:
                    assert isinstance(
                        intersection, shapely.geometry.Polygon
                    ), intersection
                    polygons.append(intersection)
    return polygonUnion(polygons)


def relativeHeadingRange(
    baseHeading, offsetL, offsetR, targetHeading, tOffsetL, tOffsetR
):
    """Lower/upper bound the possible RH between two headings with bounded disturbances."""
    if baseHeading is None or targetHeading is None:
        # heading may not be constant within cell
        return -math.pi, math.pi
    lower = normalizeAngle(baseHeading + offsetL)
    upper = normalizeAngle(baseHeading + offsetR)
    points = [lower, upper]
    if upper < lower:
        points.extend((math.pi, -math.pi))
    tLower = normalizeAngle(targetHeading + tOffsetL)
    tUpper = normalizeAngle(targetHeading + tOffsetR)
    tPoints = [tLower, tUpper]
    if tUpper < tLower:
        tPoints.extend((math.pi, -math.pi))
    rhs = [tp - p for tp in tPoints for p in points]  # TODO improve
    return min(rhs), max(rhs)


def percentagePruned(base, newBase):
    if needsSampling(base) or needsSampling(newBase):
        return None

    if (
        base.dimensionality
        and newBase.dimensionality
        and base.dimensionality == newBase.dimensionality
    ):
        ratio = newBase.size / base.size
        percent = max(0, 100 * (1.0 - ratio))
        return percent

    return None


def checkConditionedCycle(A, B):
    """Returns true if A depends on B"""
    if A is B:
        return True

    deps = set()
    unseen_deps = conditionedDeps(A)

    A = conditionedVal(A)
    B = conditionedVal(B)

    while unseen_deps:
        target_dep = unseen_deps.pop()
        new_deps = conditionedDeps(target_dep)

        if target_dep is B or any(d is B for d in new_deps):
            return True

        unseen_deps += [d for d in new_deps if d not in deps]
        deps.update(new_deps)

    return False


def conditionedDeps(samp):
    return [conditionedVal(s) for s in dependencies(conditionedVal(samp))]


def conditionedVal(samp):
    return samp._conditioned if isinstance(samp, Samplable) else samp



"""Support for hard and soft requirements."""

from abc import ABC, abstractmethod
import enum
from functools import reduce
import inspect
import itertools

import rv_ltl
import trimesh

from scenic.core.distributions import Samplable, needsSampling
from scenic.core.errors import InvalidScenarioError
from scenic.core.lazy_eval import needsLazyEvaluation
from scenic.core.propositions import Atomic, PropositionNode
import scenic.syntax.relations as relations


@enum.unique
class RequirementType(enum.Enum):
    # requirements which must hold during initial sampling
    require = "require"

    # requirements used only during simulation
    monitor = "require monitor"
    terminateWhen = "terminate when"
    terminateSimulationWhen = "terminate simulation when"

    # recorded values, which aren't requirements but are handled similarly
    record = "record"
    recordInitial = "record initial"
    recordFinal = "record final"

    @property
    def constrainsSampling(self):
        return self in (self.require,)


class PendingRequirement:
    def __init__(self, ty, condition, line, prob, name, ego):
        self.ty = ty
        self.condition = condition
        self.line = line
        self.prob = prob
        self.name = name

        # the translator wrapped the requirement in a lambda to prevent evaluation,
        # so we need to save the current values of all referenced names; we save
        # the ego object too since it can be referred to implicitly

        # condition is an instance of Proposition. Flatten to get a list of atomic propositions.
        atoms = condition.atomics()
        bindings = {}
        atomGlobals = None
        for atom in atoms:
            bindings.update(getAllGlobals(atom.closure))
            globs = atom.closure.__globals__
            if atomGlobals is not None:
                assert globs is atomGlobals
            else:
                atomGlobals = globs
        self.bindings = bindings
        self.egoObject = ego

    def compile(self, namespace, scenario, syntax=None):
        """Create a closure testing the requirement in the correct runtime state.

        While we're at it, determine whether the requirement implies any relations
        we can use for pruning, and gather all of its dependencies.
        """
        bindings, ego, line = self.bindings, self.egoObject, self.line
        condition, ty = self.condition, self.ty

        # Check whether requirement implies any relations used for pruning
        canPrune = condition.check_constrains_sampling()
        if canPrune:
            relations.inferRelationsFrom(syntax, bindings, ego, line)

        # Gather dependencies of the requirement
        deps = set()
        for value in bindings.values():
            if needsSampling(value):
                deps.add(value)
            if needsLazyEvaluation(value):
                raise InvalidScenarioError(
                    f"{ty} on line {line} uses value {value}"
                    " undefined outside of object definition"
                )

        # If this requirement contains the CanSee specifier, we will need to sample all objects
        # to meet the dependencies.
        if "CanSee" in bindings:
            deps.update(scenario.objects)

        if ego is not None:
            assert isinstance(ego, Samplable)
            deps.add(ego)

        # Construct closure
        def closure(values, monitor=None):
            # rebind any names referring to sampled objects
            # note: need to extract namespace here rather than close over value
            # from above because of https://github.com/uqfoundation/dill/issues/532
            namespace = condition.atomics()[0].closure.__globals__
            for name, value in bindings.items():
                if value in values:
                    namespace[name] = values[value]
            # rebind ego object, which can be referred to implicitly
            boundEgo = None if ego is None else values[ego]
            # evaluate requirement condition, reporting errors on the correct line
            import scenic.syntax.veneer as veneer

            with veneer.executeInRequirement(scenario, boundEgo, values):
                if monitor is None:
                    # if not temporal evaluation
                    result = condition.evaluate()
                else:
                    # if temporal evaluation
                    result = monitor.update()
                assert not needsSampling(result)
                if needsLazyEvaluation(result):
                    raise InvalidScenarioError(
                        f"{ty} on line {line} uses value"
                        " undefined outside of object definition"
                    )
            return result

        return CompiledRequirement(self, closure, deps, condition)


def getAllGlobals(req, restrictTo=None):
    """Find all names the given lambda depends on, along with their current bindings."""
    namespace = req.__globals__
    if restrictTo is not None and restrictTo is not namespace:
        return {}
    externals = inspect.getclosurevars(req)
    assert not externals.nonlocals  # TODO handle these
    globs = dict(externals.builtins)
    for name, value in externals.globals.items():
        globs[name] = value
        if inspect.isfunction(value):
            subglobs = getAllGlobals(value, restrictTo=namespace)
            for name, value in subglobs.items():
                if name in globs:
                    assert value is globs[name]
                else:
                    globs[name] = value
    return globs


class BoundRequirement:
    def __init__(self, compiledReq, sample, proposition):
        self.ty = compiledReq.ty
        self.closure = compiledReq.closure
        self.line = compiledReq.line
        self.name = compiledReq.name
        self.sample = sample
        self.compiledReq = compiledReq
        self.proposition = proposition

    def isTrue(self):
        return self.value()

    def value(self):
        one_time_monitor = self.proposition.create_monitor()
        return self.closure(self.sample, one_time_monitor)

    def evaluate(self):
        return self.closure(self.sample)

    def __str__(self):
        if self.name:
            return self.name
        else:
            return f'"{self.ty.value}" on line {self.line}'

    def toMonitor(self):
        return MonitorRequirement(self.compiledReq, self.sample, self.proposition)


class MonitorRequirement(BoundRequirement):
    """MonitorRequirement is a BoundRequirement with temporal proposition monitor"""

    def __init__(self, compiledReq, sample, proposition):
        super().__init__(compiledReq, sample, proposition)
        self.monitor = self.proposition.create_monitor()
        self.lastValue = rv_ltl.B4.TRUE

    def value(self):
        self.lastValue = self.closure(self.sample, self.monitor)
        return self.lastValue


class DynamicRequirement:
    def __init__(self, ty, condition, line, name=None):
        self.ty = ty
        self.line = line
        self.name = name

        import scenic.syntax.veneer as veneer

        scenario = veneer.currentScenario

        def closure(monitor=None):
            with veneer.executeInScenario(scenario):
                if monitor is None:
                    result = self.condition.evaluate()
                else:
                    result = monitor.update()
                return result

        self.closure = closure
        self.condition = condition

    def isTrue(self):
        return self.value()

    def value(self):
        return self.closure()

    def __str__(self):
        if self.name:
            return self.name
        else:
            return f'"{self.ty.value}" on line {self.line}'

    def toMonitor(self):
        return DynamicMonitorRequirement(
            self.closure, self.condition, self.line, self.name
        )


class DynamicMonitorRequirement:
    def __init__(self, closure, condition, line, name):
        self.line = line
        self.closure = closure
        self.name = name
        self.condition = condition
        self.monitor = self.condition.create_monitor()
        self.lastValue = rv_ltl.B4.TRUE

    def isTrue(self):
        return self.value()

    def value(self):
        self.lastValue = self.closure(self.monitor)
        return self.lastValue

    def evaluate(self):
        return self.closure()

    def __str__(self):
        if self.name:
            return self.name
        else:
            return f'"{self.ty.value}" on line {self.line}'


## Builtin Requirements
class SamplingRequirement(ABC):
    """A requirement to be checked to validate a sample.

    Args:
        optional: Whether or not this requirement must be
            checked to validate the sample. Optional samples
            can be checked, and if ``False`` imply that the
            sample is invalid, but do not need to be checked
            if all non-optional requirements are satisfied.
    """

    def __init__(self, optional):
        self.optional = optional
        self.active = True

    @abstractmethod
    def falsifiedByInner(self, sample):
        """Returns False if the requirement is falsifed, True otherwise"""

    def falsifiedBy(self, sample):
        assert self.active
        return self.falsifiedByInner(sample)

    @property
    @abstractmethod
    def violationMsg(self):
        """Message to be printed if the requirement is violated"""
        pass


class IntersectionRequirement(SamplingRequirement):
    def __init__(self, objA, objB, optional=False):
        super().__init__(optional=optional)
        self.objA = objA
        self.objB = objB

    def falsifiedByInner(self, sample):
        objA = sample[self.objA]
        objB = sample[self.objB]
        if objA.allowCollisions or objB.allowCollisions:
            return False
        return objA.intersects(objB)

    @property
    def violationMsg(self):
        return f"Intersection violation: {self.objA} intersects {self.objB}"


class BlanketCollisionRequirement(SamplingRequirement):
    def __init__(self, objects, optional=True):
        super().__init__(optional=optional)
        self.objects = objects
        self._collidingObjects = None

    def falsifiedByInner(self, sample):
        objects = tuple(sample[obj] for obj in self.objects)
        cm = trimesh.collision.CollisionManager()
        for i, obj in enumerate(objects):
            if not obj.allowCollisions:
                cm.add_object(str(i), obj.occupiedSpace.mesh)
        collision, names = cm.in_collision_internal(return_names=True)

        if collision:
            self._collidingObjects = tuple(sorted(names))

        return collision

    @property
    def violationMsg(self):
        assert self._collidingObjects is not None
        objA_index, objB_index = map(int, self._collidingObjects[0])
        objA, objB = self.objects[objA_index], self.objects[objB_index]
        return f"Intersection violation: {objA} intersects {objB}"


class ContainmentRequirement(SamplingRequirement):
    def __init__(self, obj, container, optional=False):
        super().__init__(optional=optional)
        self.obj = obj
        self.container = container

    def falsifiedByInner(self, sample):
        obj = sample[self.obj]
        container = sample[self.container]
        return not container.containsObject(obj)

    @property
    def violationMsg(self):
        return f"Containment violation: {self.obj} is not contained in its container"


class VisibilityRequirement(SamplingRequirement):
    def __init__(self, source, target, objects, optional=False):
        super().__init__(optional=optional)
        self.source = source
        self.target = target
        self.potential_occluders = tuple(
            obj for obj in objects if obj is not self.source and obj is not self.target
        )

    def falsifiedByInner(self, sample):
        source = sample[self.source]
        target = sample[self.target]
        potential_occluders = tuple(sample[obj] for obj in self.potential_occluders)
        occluders = tuple(obj for obj in potential_occluders if obj.occluding)
        return not source.canSee(target, occludingObjects=occluders)

    @property
    def violationMsg(self):
        return f"Visibility violation: {self.target} is not visible from {self.source}"


class NonVisibilityRequirement(VisibilityRequirement):
    def falsifiedByInner(self, sample):
        return not super().falsifiedByInner(sample)

    @property
    def violationMsg(self):
        return f"Non-visibility violation: {self.target} is visible from {self.source}"


class CompiledRequirement(SamplingRequirement):
    def __init__(self, pendingReq, closure, dependencies, proposition):
        super().__init__(optional=False)
        self.ty = pendingReq.ty
        self.closure = closure
        self.line = pendingReq.line
        self.name = pendingReq.name
        self.prob = pendingReq.prob
        self.dependencies = dependencies
        self.proposition = proposition

    @property
    def constrainsSampling(self):
        return self.ty.constrainsSampling

    def falsifiedByInner(self, sample):
        one_time_monitor = self.proposition.create_monitor()
        return self.closure(sample, one_time_monitor) == rv_ltl.B4.FALSE

    def __str__(self):
        if self.name:
            return self.name
        else:
            return f'"{self.ty.value}" on line {self.line}'

    @property
    def violationMsg(self):
        return f"User requirement violation: {self}"



"""Parser for WBT files using ANTLR.

The ANTLR parser itself, consisting of the *WBTLexer.py*, *WBTParser.py*, and *WBTVisitor.py* files, is autogenerated from *WBT.g4*.
"""

import io
import sys
import time

import antlr4
from antlr4.error.ErrorListener import ErrorListener
import numpy as np

from scenic.simulators.webots.WBTLexer import WBTLexer
from scenic.simulators.webots.WBTParser import WBTParser
from scenic.simulators.webots.WBTVisitor import WBTVisitor

### Setup

## Classes representing various types of VRML nodes


class Node:
    """A generic VRML node"""

    def __init__(self, nodeType, attrs):
        self.nodeType = nodeType
        self.attrs = attrs


## Classes for parsing with ANTLR

# Error handling


class ErrorReporter(ErrorListener):
    """ANTLR listener for reporting parse errors"""

    def syntaxError(self, recognizer, offendingSymbol, line, column, msg, e):
        raise Exception(f"Parse error at {line}:{column}: {msg}")


# Walking the parse tree and extracting values


class Evaluator(WBTVisitor):
    """Constructs an object representing the given value from the parse tree"""

    def __init__(self, nodeClasses):
        super().__init__()
        # Dictionary indicating which types of nodes have specialized classes
        self.nodeClasses = nodeClasses

    def visitNode(self, node):
        nodeType = node.Identifier().getText()
        attrs = self.visitNodeBody(node.nodeBody())
        if nodeType in self.nodeClasses:
            nodeClass = self.nodeClasses[nodeType]
            nodeObj = nodeClass(attrs)
        else:
            nodeObj = Node(nodeType, attrs)
        return nodeObj

    def visitNodeBody(self, body):
        return {
            attr.Identifier().getText(): self.visitValue(attr.value())
            for attr in body.attribute()
        }

    def visitValue(self, val):
        return self.visit(val.getChild(0))

    def visitVector(self, vec):
        return np.array([float(n.getText()) for n in vec.Number()])

    def visitVectorWithNewlines(self, vec):
        return self.visitVector(vec)

    def visitString(self, string):
        return string.getText()[1:-1]  # strip off quotes

    def visitArray(self, array):
        return [
            self.visit(val)
            for val in array.children
            if not isinstance(val, antlr4.tree.Tree.TerminalNode)
        ]

    def visitBoolean(self, boolean):
        return boolean.getText() == "TRUE"


def parse(path):
    """Parse a world from a WBT file"""
    # Open file
    charStream = antlr4.FileStream(path)

    # Set up lexer and parser
    errorListener = ErrorReporter()
    lexer = WBTLexer(charStream)
    lexer.addErrorListener(errorListener)
    tokenStream = antlr4.CommonTokenStream(lexer)
    parser = WBTParser(tokenStream)
    parser.addErrorListener(errorListener)

    # Parse file
    return parser.world()


def findNodeTypesIn(types, world, nodeClasses={}):
    """Find all nodes of the given types in a world"""
    evaluator = Evaluator(nodeClasses)
    types = tuple(types)
    instances = {ty: [] for ty in types}
    for node in world.children:
        if isinstance(node, WBTParser.DefnContext):
            node = node.node()
        nodeType = node.Identifier().getText()
        if nodeType in instances:
            val = evaluator.visitNode(node)
            instances[nodeType].append(val)
    return tuple(instances[ty] for ty in types)



"""Scenario and scene objects."""

import dataclasses
import io
import itertools
import random
import sys
import time

import numpy
import trimesh

import scenic
from scenic.core.distributions import (
    ConstantSamplable,
    RejectionException,
    Samplable,
    distributionFunction,
    needsSampling,
)
from scenic.core.dynamics.behaviors import Behavior, Monitor
from scenic.core.errors import InvalidScenarioError, optionallyDebugRejection
from scenic.core.external_params import ExternalSampler
from scenic.core.lazy_eval import needsLazyEvaluation
from scenic.core.regions import (
    AllRegion,
    EmptyRegion,
    PointInRegionDistribution,
    convertToFootprint,
)
from scenic.core.requirements import (
    BlanketCollisionRequirement,
    BoundRequirement,
    ContainmentRequirement,
    IntersectionRequirement,
    NonVisibilityRequirement,
    VisibilityRequirement,
)
from scenic.core.sample_checking import BasicChecker, WeightedAcceptanceChecker
from scenic.core.serialization import Serializer, dumpAsScenicCode
from scenic.core.vectors import Vector

# Global params

INITIAL_COLLISION_CHECK = True

# Pickling support


class _ScenarioPickleMixin:
    def __getstate__(self):
        # Start the pickle with an object storing our compile options and activating
        # the veneer with them; this will ensure they are consistent during import of
        # any needed Scenic modules (which might have been purged earlier, and in any
        # case won't already exist during unpickling). Similarly, tack a dummy object
        # on the end of the pickle which will deactivate the veneer and clean up.
        oldModules = []
        options = dataclasses.replace(
            self.compileOptions,
            paramOverrides=self.params,  # save all params, not just those from --param
        )
        elements = (
            _Activator(options, oldModules),
            self.__dict__,
            _Deactivator(oldModules),
        )
        return elements

    def __setstate__(self, state):
        self.__dict__.update(state[1])


class _Activator:
    def __init__(self, compileOptions, oldModules):
        self.compileOptions = compileOptions
        # Save all modules already imported prior to pickling
        oldModules.extend(sys.modules.keys())

    def activate(self):
        import scenic.syntax.veneer as veneer

        assert not veneer.isActive()
        veneer.activate(self.compileOptions)

    def __getstate__(self):
        # Step 1 (during pickling)
        self.activate()
        return self.__dict__

    def __setstate__(self, state):
        # Step 3 (during unpickling)
        self.__dict__.update(state)
        self.activate()


class _Deactivator:
    def __init__(self, oldModules):
        self.oldModules = oldModules

    def deactivate(self):
        from scenic.syntax.translator import purgeModulesUnsafeToCache
        import scenic.syntax.veneer as veneer

        veneer.deactivate()
        assert not veneer.isActive(), "nested pickle of Scene/Scenario"
        # Purge Scenic modules imported during pickling
        purgeModulesUnsafeToCache(self.oldModules)

    def __getstate__(self):
        # Step 2 (during pickling)
        self.deactivate()
        return self.__dict__

    def __setstate__(self, state):
        # Step 4 (during unpickling)
        self.__dict__.update(state)
        self.deactivate()


# Scenes and scenarios


class Scene(_ScenarioPickleMixin):
    """Scene()

    A scene generated from a Scenic scenario.

    To run a dynamic simulation from a scene, create an instance of `Simulator` for the
    simulator you want to use, and pass the scene to its `simulate` method.

    Attributes:
        objects (tuple of :obj:`~scenic.core.object_types.Object`): All objects in the
          scene. The ``ego`` object is first, if there is one.
        egoObject (:obj:`~scenic.core.object_types.Object` or `None`): The ``ego`` object, if any.
        params (dict): Dictionary mapping the name of each global parameter to its value.
        workspace (:obj:`~scenic.core.workspaces.Workspace`): The :term:`workspace` for the scenario.

    .. versionchanged:: 3.0

        The ``egoObject`` attribute can now be `None`.
    """

    def __init__(
        self,
        workspace,
        objects,
        egoObject,
        params,
        temporalReqs=(),
        terminationConds=(),
        termSimulationConds=(),
        recordedExprs=(),
        recordedInitialExprs=(),
        recordedFinalExprs=(),
        monitors=(),
        behaviorNamespaces={},
        dynamicScenario=None,
        sample={},
        compileOptions={},
    ):
        self.workspace = workspace
        self.objects = tuple(objects)
        self.egoObject = egoObject
        self.params = params
        self.temporalRequirements = tuple(temporalReqs)
        self.terminationConditions = tuple(terminationConds)
        self.terminateSimulationConditions = tuple(termSimulationConds)
        self.recordedExprs = tuple(recordedExprs)
        self.recordedInitialExprs = tuple(recordedInitialExprs)
        self.recordedFinalExprs = tuple(recordedFinalExprs)
        self.monitors = tuple(monitors)
        self.behaviorNamespaces = behaviorNamespaces
        self.dynamicScenario = dynamicScenario
        self.sample = sample
        self.compileOptions = compileOptions

    def dumpAsScenicCode(self, stream=sys.stdout):
        """Dump Scenic code reproducing this scene to the given stream.

        For non-human-readable but complete serialization of scenes see
        `Scenario.sceneToBytes` and `Scenario.sceneFromBytes`.

        .. note::

            This function does not currently reproduce parts of the original Scenic
            program defining behaviors, functions, etc. used in the scene. Also, if
            the scene involves any user-defined types, they must provide a suitable
            :obj:`~object.__repr__` for this function to print them properly.

        Args:
            stream (:term:`text file`): Where to print the code (default `sys.stdout`).
        """
        for name, value in self.params.items():
            if str.isidentifier(name):
                stream.write(f"param {name} = ")
            else:
                stream.write(f'param "{name}" = ')
            dumpAsScenicCode(value, stream)
            stream.write("\n")
        stream.write("ego = ")
        for obj in self.objects:
            dumpAsScenicCode(obj, stream)
            stream.write("\n")

    def show(self, axes=True):
        return self.show3D(axes=axes)

    def show3D(self, axes):
        """Render a 3D schematic of the scene for debugging."""
        import trimesh

        # Create a new trimesh scene to contain meshes
        render_scene = trimesh.scene.Scene()

        # display map
        self.workspace.show3D(render_scene)
        # draw objects
        for obj in self.objects:
            obj.show3D(render_scene, highlight=(obj is self.egoObject))

        # If nothing else is in the viewer, add some constructs
        # to avoid a crash
        if render_scene.is_empty:
            render_scene.add_geometry(
                trimesh.points.PointCloud(
                    [(0.1, 0, 0), (0, 0.1, 0), (0, 0, 0.1)], colors=[255, 255, 255, 0]
                )
            )

        flags = dict()
        if axes:
            flags["axis"] = "world"

        return render_scene.show(flags=flags)

    def show2D(self, zoom=None, block=True):
        """Render a 2D schematic of the scene for debugging."""
        import matplotlib.pyplot as plt

        plt.gca().set_aspect("equal")
        # display map
        self.workspace.show2D(plt)
        # draw objects
        for obj in self.objects:
            obj.show2D(self.workspace, plt, highlight=(obj is self.egoObject))
        # zoom in if requested
        if zoom:
            self.workspace.zoomAround(plt, self.objects, expansion=zoom)
        plt.show(block=block)


class Scenario(_ScenarioPickleMixin):
    """Scenario()

    A compiled Scenic scenario, from which scenes can be sampled.
    """

    def __init__(
        self,
        workspace,
        simulator,
        instances,
        objects,
        egoObject,
        params,
        externalParams,
        requirements,
        requirementDeps,
        monitors,
        behaviorNamespaces,
        dynamicScenario,
        astHash,
        compileOptions,
    ):
        self.workspace = workspace
        self.simulator = simulator  # simulator for dynamic scenarios
        # make ego the first object, while otherwise preserving order
        ordered = []
        for obj in objects:
            if obj is not egoObject:
                ordered.append(obj)
        assert set(objects).issubset(set(instances))
        self._instances = tuple(instances)
        self.objects = (egoObject,) + tuple(ordered) if egoObject else tuple(ordered)
        self.egoObject = egoObject
        self.params = dict(params)
        self.externalParams = tuple(externalParams)
        self.externalSampler = ExternalSampler.forParameters(
            self.externalParams, self.params
        )
        self.monitors = tuple(monitors)
        self.behaviorNamespaces = behaviorNamespaces
        self.dynamicScenario = dynamicScenario
        self.astHash = astHash
        self.compileOptions = compileOptions

        staticReqs, alwaysReqs, terminationConds = [], [], []
        self.requirements = tuple(dynamicScenario._requirements)  # TODO clean up
        self.terminationConditions = tuple(dynamicScenario._terminationConditions)
        self.terminateSimulationConditions = tuple(
            dynamicScenario._terminateSimulationConditions
        )
        self.userRequirements = self.requirements
        assert all(req.constrainsSampling for req in self.userRequirements)
        self.recordedExprs = tuple(dynamicScenario._recordedExprs)
        self.recordedInitialExprs = tuple(dynamicScenario._recordedInitialExprs)
        self.recordedFinalExprs = tuple(dynamicScenario._recordedFinalExprs)
        # dependencies must use fixed order for reproducibility
        paramDeps = tuple(p for p in self.params.values() if isinstance(p, Samplable))
        behaviorDeps = []
        for namespace in self.behaviorNamespaces.values():
            for value in namespace.values():
                if isinstance(value, Samplable):
                    behaviorDeps.append(value)
        self.dependencies = (
            self._instances + paramDeps + tuple(requirementDeps) + tuple(behaviorDeps)
        )

        # Setup the default checker
        self.defaultRequirements = self.generateDefaultRequirements()
        self.setSampleChecker(WeightedAcceptanceChecker(bufferSize=100))

    def setSampleChecker(self, checker):
        self.checker = checker
        self.checker.setRequirements(self.defaultRequirements + self.userRequirements)

    def containerOfObject(self, obj):
        if hasattr(obj, "regionContainedIn") and obj.regionContainedIn is not None:
            return obj.regionContainedIn
        else:
            return convertToFootprint(self.workspace.region)

    def validate(self):
        """Make some simple static checks for inconsistent built-in requirements.

        :meta private:
        """
        objects = self.objects
        ego = self.egoObject
        staticVisibility = ego and not needsSampling(ego.visibleRegion)
        staticBounds = [obj._hasStaticBounds for obj in objects]
        for i in range(len(objects)):
            oi = objects[i]
            container = self.containerOfObject(oi)
            # Trivial case where container is empty
            if isinstance(container, EmptyRegion):
                raise InvalidScenarioError(f"Container region of {oi} is empty")
            # Ensure we are not sampling position from AllRegion
            if isinstance(
                oi.position._conditioned, PointInRegionDistribution
            ) and isinstance(oi.position._conditioned.region, AllRegion):
                if oi.position.tag == "visible":
                    raise InvalidScenarioError(
                        f"Object {oi} uses the visible specifier to specify position, but it lacks enough information to do so."
                        f" The simplest solution to this is to define a workspace or specify position in some other fashion."
                    )
                else:
                    raise InvalidScenarioError(
                        f"Object {oi} has position sampled from everywhere."
                    )
            # skip objects with unknown positions or bounding boxes
            if not staticBounds[i]:
                continue
            # Require object to be contained in the workspace/valid region
            if not needsSampling(container) and not container.containsObject(oi):
                raise InvalidScenarioError(
                    f"Object at {oi.position} does not fit in container"
                )
            # Require object to be visible from the ego object
            if staticVisibility and oi.requireVisible is True and oi is not ego:
                if not ego.canSee(oi):
                    raise InvalidScenarioError(
                        f"Object at {oi.position} is not visible from ego"
                    )
            if not needsSampling(oi.allowCollisions) and not oi.allowCollisions:
                # Require object to not intersect another object
                for j in range(i):
                    oj = objects[j]
                    if oj.allowCollisions or not staticBounds[j]:
                        continue
                    if oi.intersects(oj):
                        raise InvalidScenarioError(
                            f"{oi} at {oi.position} intersects" f" {oj} at {oj.position}"
                        )

    def generate(self, maxIterations=2000, verbosity=0, feedback=None):
        """Sample a `Scene` from this scenario.

        For a description of how scene generation is done, see `scene generation`.

        Args:
            maxIterations (int): Maximum number of rejection sampling iterations.
            verbosity (int): Verbosity level.
            feedback (float): Feedback to pass to external samplers doing active sampling.
                See :mod:`scenic.core.external_params`.

        Returns:
            A pair with the sampled `Scene` and the number of iterations used.

        Raises:
            `RejectionException`: if no valid sample is found in **maxIterations** iterations.
        """
        scenes, iterations = self.generateBatch(1, maxIterations, verbosity, feedback)
        return scenes[0], iterations

    def generateBatch(
        self, numScenes, maxIterations=float("inf"), verbosity=0, feedback=None
    ):
        """Sample several `Scene` objects from this scenario.

        For a description of how scene generation is done, see `scene generation`.

        Args:
            numScenes (int): Number of scenes to generate.
            maxIterations (int): Maximum number of rejection sampling iterations (over all scenes).
            verbosity (int): Verbosity level.
            feedback (float): Feedback to pass to external samplers doing active sampling.
                See :mod:`scenic.core.external_params`.

        Returns:
            A pair with a list of the sampled `Scene` objects and the total number
            of iterations used.

        Raises:
            `RejectionException`: if not enough valid samples are found in **maxIterations** iterations.
        """
        totalIterations = 0
        scenes = []

        for _ in range(numScenes):
            try:
                remainingIts = maxIterations - totalIterations
                scene, iterations = self._generateInner(remainingIts, verbosity, feedback)
                scenes.append(scene)
                totalIterations += iterations
            except RejectionException:
                raise RejectionException(
                    f"failed to generate scenario in {maxIterations} iterations"
                )

        return scenes, totalIterations

    def _generateInner(self, maxIterations, verbosity, feedback):
        # choose which custom requirements will be enforced for this sample
        for req in self.userRequirements:
            if random.random() <= req.prob:
                req.active = True
            else:
                req.active = False

        # do rejection sampling until requirements are satisfied
        rejection = True
        iterations = 0
        while rejection is not None:
            if iterations > 0:  # rejected the last sample
                if verbosity >= 2:
                    print(f"  Rejected sample {iterations} because of {rejection}")
                if self.externalSampler is not None:
                    feedback = self.externalSampler.rejectionFeedback
            if iterations >= maxIterations:
                raise RejectionException(
                    f"failed to generate scenario in {iterations} iterations"
                )
            iterations += 1
            try:
                if self.externalSampler is not None:
                    self.externalSampler.sample(feedback)
                sample = Samplable.sampleAll(self.dependencies)
            except RejectionException as e:
                optionallyDebugRejection(e)
                rejection = e
                continue
            rejection = None

            # Ensure nothing else is lazy
            for obj in self.objects:
                sampledObj = sample[obj]
                assert not needsSampling(sampledObj)

            # Check validity of sample, storing state so that
            # checker heuristics don't affect determinism
            rand_state, np_state = random.getstate(), numpy.random.get_state()
            rejection = self.checker.checkRequirements(sample)
            random.setstate(rand_state)
            numpy.random.set_state(np_state)

            if rejection is not None:
                optionallyDebugRejection()

        # obtained a valid sample; assemble a scene from it
        scene = self._makeSceneFromSample(sample)
        return scene, iterations

    def generateDefaultRequirements(self):
        requirements = []

        ## Optional Requirements ##
        # Any collision indicates an intersection
        requirements.append(BlanketCollisionRequirement(self.objects))

        ## Mandatory Requirements ##
        # Pairwise object intersection
        colliding_objects = (
            obj
            for obj in self.objects
            if needsSampling(obj.allowCollisions) or not obj.allowCollisions
        )
        for objA, objB in itertools.combinations(colliding_objects, 2):
            requirements.append(IntersectionRequirement(objA, objB))

        # Object containment
        for obj in self.objects:
            container = self.containerOfObject(obj)
            if not isinstance(container, AllRegion):
                requirements.append(ContainmentRequirement(obj, container))

        # Observing entity visibility
        possible_occluders = filter(
            lambda x: (needsSampling(x.occluding) or x.occluding), self.objects
        )
        for obj in filter(lambda x: x._observingEntity is not None, self._instances):
            requirements.append(
                VisibilityRequirement(obj._observingEntity, obj, possible_occluders)
            )

        # Observing entity non visibility
        for obj in filter(lambda x: x._nonObservingEntity is not None, self._instances):
            requirements.append(
                NonVisibilityRequirement(obj._nonObservingEntity, obj, possible_occluders)
            )

        # Visibility from the ego
        for obj in filter(
            lambda x: x.requireVisible and x is not self.egoObject, self.objects
        ):
            if not self.egoObject:
                raise InvalidScenarioError(
                    "requireVisible set to true but no ego is defined"
                )
            requirements.append(VisibilityRequirement(self.egoObject, obj, self.objects))

        return tuple(requirements)

    def _makeSceneFromSample(self, sample):
        sampledObjects = tuple(sample[obj] for obj in self.objects)
        ego = sample[self.egoObject]
        sampledParams = {}
        for param, value in self.params.items():
            sampledValue = sample[value]
            assert not needsLazyEvaluation(sampledValue)
            sampledParams[param] = sampledValue
        sampledNamespaces = {}
        for modName, namespace in self.behaviorNamespaces.items():
            sampledNamespace = {name: sample[value] for name, value in namespace.items()}
            sampledNamespaces[modName] = (namespace, sampledNamespace, namespace.copy())
        temporalReqs = (
            BoundRequirement(req, sample, req.proposition) for req in self.requirements
        )
        monitors = []
        for req in self.monitors:
            breq = BoundRequirement(req, sample, None)
            monitor = breq.evaluate()
            if not isinstance(monitor, Monitor):
                raise TypeError(
                    f'"require monitor X" with X not a monitor on line {breq.line}'
                )
            monitors.append(monitor)
        terminationConds = (
            BoundRequirement(req, sample, req.proposition)
            for req in self.terminationConditions
        )
        termSimulationConds = (
            BoundRequirement(req, sample, req.proposition)
            for req in self.terminateSimulationConditions
        )
        recordedExprs = (
            BoundRequirement(req, sample, req.proposition) for req in self.recordedExprs
        )
        recordedInitialExprs = (
            BoundRequirement(req, sample, req.proposition)
            for req in self.recordedInitialExprs
        )
        recordedFinalExprs = (
            BoundRequirement(req, sample, req.proposition)
            for req in self.recordedFinalExprs
        )
        scene = Scene(
            self.workspace,
            sampledObjects,
            ego,
            sampledParams,
            temporalReqs,
            terminationConds,
            termSimulationConds,
            recordedExprs,
            recordedInitialExprs,
            recordedFinalExprs,
            monitors,
            sampledNamespaces,
            self.dynamicScenario,
            sample,
            self.compileOptions,
        )
        return scene

    def resetExternalSampler(self):
        """Reset the scenario's external sampler, if any.

        If the Python random seed is reset before calling this function, this
        should cause the sequence of generated scenes to be deterministic.
        """
        self.externalSampler = ExternalSampler.forParameters(
            self.externalParams, self.params
        )

    def conditionOn(self, scene=None, objects=(), params={}):
        """Condition the scenario on particular values for some objects or parameters.

        This method changes the distribution of the scenario and should be used with
        care: it does not attempt to check that the new distribution is equivalent to the
        old one or that it has nonzero probability of satisfying the scenario's
        requirements.

        For example, to sample object #5 in the scenario once and then leave it fixed in
        all subsequent samples::

            sceneA, _ = scenario.generate()
            scenario.conditionOn(scene=sceneA, objects=(5,))
            sceneB, _ = scenario.generate()     # will have the same object 5 as sceneA

        Args:
            scene (Scene): Scene from which to take values for the given **objects**,
                if any.
            objects: Sequence of indices specifying which objects in this scenario should
                be conditioned on the corresponding objects in **scene** (i.e. those with
                the same index in the list of objects).
            params (dict): Dictionary of global parameters to condition and their new
                values (which may be constants or distributions).
        """
        assert objects or params
        assert bool(scene) == bool(objects)
        if scene:
            assert len(self.objects) == len(scene.objects)
        for i in objects:
            assert i < len(self.objects)
            self.objects[i].conditionTo(scene.objects[i])
        for param, newVal in params.items():
            curVal = self.params[param]
            if isinstance(curVal, Samplable):
                if not isinstance(newVal, Samplable):
                    newVal = ConstantSamplable(newVal)
                curVal.conditionTo(newVal)
            else:
                self.params[param] = newVal

    def getSimulator(self):
        if self.simulator is None:
            raise RuntimeError("scenario does not specify a simulator")
        import scenic.syntax.veneer as veneer

        return veneer.instantiateSimulator(self.simulator, self.params)

    def sceneToBytes(self, scene, allowPickle=False):
        """Encode a `Scene` sampled from this scenario to a `bytes` object.

        The serialized scene may be reconstituted with `sceneFromBytes`. The format used
        is suitable for long-term storage of scenes, although it is not guaranteed to be
        compatible across major versions of Scenic. For further discussion and usage
        examples, see :ref:`serialization`.

        Raises:
            SerializationError: if the scene could not be properly encoded. This should
                not happen unless your scenario includes a user-defined `Distribution`
                subclass with an unusual value type. If you get this exception, see the
                documentation for the internal class `Serializer` for solutions.
        """
        ser = Serializer(allowPickle=allowPickle)
        ser.writeScene(self, scene)
        return ser.getBytes()

    def sceneFromBytes(self, data, verify=True, allowPickle=False):
        """Decode a `Scene` serialized with `sceneToBytes`.

        Args:
            data (bytes): Encoding of a `Scene` sampled from this scenario.
            verify (bool): If true (the default), raise an exception if the scene
                appears to have been generated from a different scenario (meaning
                it will almost certainly not decode correctly).
            allowPickle (bool): Enable using `pickle` to deserialize custom object
                types. False by default because it allows malicious data to trigger
                arbitrary code execution (see the `pickle` documentation). Use this
                option only if you trust the source of the data and it is not practical
                to implement serialization for the datatypes you need.

        Raises:
            SerializationError: if the scene could not be properly decoded.
        """
        ser = Serializer(data, allowPickle=allowPickle)
        return ser.readScene(self, verify=verify)

    def simulationToBytes(self, simulation, allowPickle=False):
        """Encode a `Simulation` sampled from this scenario to a `bytes` object.

        The serialized simulation may be replayed with `simulationFromBytes`. As with
        `sceneToBytes`, the format used is suitable for long-term storage but is not
        guaranteed to be compatible across major versions of Scenic.

        Raises:
            SerializationError: if the simulation could not be properly encoded. This should
                not happen unless your scenario includes a user-defined `Distribution`
                subclass with an unusual value type. If you get this exception, see the
                documentation for the internal class `Serializer` for solutions.

        .. note::

            The returned data encodes both the scene comprising the initial condition for the
            simulation and the simulation itself. If you will be running many simulations
            starting from the same scene, you can save space by separately encoding the scene
            and the various simulations: use `sceneToBytes` and `Simulation.getReplay` for
            encoding, and the **replay** argument of `Simulator.simulate` for decoding.
        """
        sceneData = self.sceneToBytes(simulation.scene)
        replay = simulation.getReplay()
        return sceneData + replay

    def simulationFromBytes(
        self, data, simulator, *, verify=True, allowPickle=False, **kwargs
    ):
        """Replay a `Simulation` serialized with `simulationToBytes`.

        Args:
            data (bytes): Encoding of a `Simulation` sampled from this scenario.
            simulator (Simulator): Simulator in which to run the simulation. Using
                a different simulator configuration than that used for the original
                simulation may cause errors or unexpected behavior. If you need to do
                this, see the **enableDivergenceCheck** option of `Simulator.simulate`.
            verify (bool): As in `sceneFromBytes`.
            allowPickle (bool): As in `sceneFromBytes`.
            kwargs: All additional keyword arguments are passed through to the simulator;
                see `Simulator.simulate` for the available configuration options.

        Returns:
            A `Simulation` object representing the completed simulation.

        Raises:
            SerializationError: if the simulation could not be properly decoded.
            DivergenceError: if the replayed simulation has diverged from the original
                (requires the original to have been run with divergence-checking support;
                see `Simulator.simulate`).
        """
        if not isinstance(data, io.BufferedIOBase):
            data = io.BytesIO(data)
        scene = self.sceneFromBytes(data, verify=verify, allowPickle=allowPickle)
        return simulator.simulate(scene, replay=data, **kwargs)



"""Workspaces."""

import numpy as np
import trimesh

from scenic.core.distributions import needsSampling
from scenic.core.errors import InvalidScenarioError
from scenic.core.geometry import findMinMax
from scenic.core.regions import (
    MeshSurfaceRegion,
    MeshVolumeRegion,
    PolygonalRegion,
    Region,
    everywhere,
)


class Workspace(Region):
    """A :term:`workspace` describing the fixed world of a scenario.

    Args:
        region (Region): The region defining the extent of the workspace
          (default `everywhere`).
    """

    def __init__(self, region=everywhere):
        if needsSampling(region):
            raise InvalidScenarioError("workspace region must be fixed")
        super().__init__("workspace", orientation=region.orientation)

        self.region = region

    def show3D(self, viewer):
        """Render a schematic of the workspace (in 3D) for debugging"""
        if isinstance(
            self.region, (MeshVolumeRegion, MeshSurfaceRegion, PolygonalRegion)
        ):
            if isinstance(self.region, (MeshVolumeRegion, MeshSurfaceRegion)):
                workspace_mesh = self.region.mesh
            else:
                workspace_mesh = self.region.footprint.boundFootprint(
                    centerZ=self.region.z, height=0.0001
                ).mesh
            # We can render this workspace as the wireframe of a mesh.
            # Filter out meshes that are close enough to planar
            edges = workspace_mesh.face_adjacency_edges[
                workspace_mesh.face_adjacency_angles > np.radians(0.1)
            ]
            vertices = workspace_mesh.vertices

            edge_path = trimesh.path.Path3D(
                **trimesh.path.exchange.misc.edges_to_path(edges, vertices)
            )

            viewer.add_geometry(edge_path)

    def show2D(self, plt):
        """Render a schematic of the workspace (in 2D) for debugging"""
        try:
            aabb = self.region.AABB
        except (NotImplementedError, TypeError):  # unbounded Regions don't support this
            return
        ((xmin, ymin), (xmax, ymax), _) = aabb
        plt.xlim(xmin, xmax)
        plt.ylim(ymin, ymax)

    def zoomAround(self, plt, objects, expansion=1):
        """Zoom the schematic around the specified objects"""
        if not objects:
            return
        positions = (self.scenicToSchematicCoords(obj.position) for obj in objects)
        x, y = zip(*positions)
        minx, maxx = findMinMax(x)
        miny, maxy = findMinMax(y)
        sx = expansion * max(self.minimumZoomSize, 2 * (maxx - minx))
        sy = expansion * max(self.minimumZoomSize, 2 * (maxy - miny))
        s = max(sx, sy) / 2.0
        s += max(max(obj.width, obj.length) for obj in objects)  # TODO improve
        cx = (maxx + minx) / 2.0
        cy = (maxy + miny) / 2.0
        plt.xlim(cx - s, cx + s)
        plt.ylim(cy - s, cy + s)

    @property
    def minimumZoomSize(self):
        return 0

    def scenicToSchematicCoords(self, coords):
        """Convert Scenic coordinates to those used for schematic rendering."""
        return coords[:2]

    def uniformPointInner(self):
        return self.region.uniformPointInner()

    def intersect(self, other, triedReversed=False):
        return self.region.intersect(other, triedReversed)

    def intersects(self, other):
        return self.region.intersects(other)

    def difference(self, other):
        return self.region.difference(other)

    def union(self, other, triedReversed=False):
        return self.region.union(other, triedReversed)

    def containsPoint(self, point):
        return self.region.containsPoint(point)

    def containsObject(self, obj):
        return self.region.containsObject(obj)

    def containsRegionInner(self, reg, tolerance):
        return self.region.containsRegionInner(reg, tolerance)

    def distanceTo(self, point):
        return self.region.distanceTo(point)

    def projectVector(self, point, onDirection):
        raise self.region.projectVector(point, onDirection)

    @property
    def AABB(self):
        return self.region.AABB

    @property
    def dimensionality(self):
        return self.region.dimensionality

    @property
    def size(self):
        return self.region.size

    def __repr__(self):
        return f"Workspace({self.region!r})"

    def __eq__(self, other):
        if type(other) is not Workspace:
            return NotImplemented
        return other.region == self.region

    def __hash__(self):
        return hash(self.region)



"""CARLA blueprints for cars, pedestrians, etc."""

#: Mapping from current names of blueprints to ones in old CARLA versions.
#:
#: We provide a tuple of old names in case they change more than once.
oldBlueprintNames = {
    "vehicle.dodge.charger_police": ("vehicle.dodge_charger.police",),
    "vehicle.lincoln.mkz_2017": ("vehicle.lincoln.mkz2017",),
    "vehicle.mercedes.coupe": ("vehicle.mercedes-benz.coupe",),
    "vehicle.mini.cooper_s": ("vehicle.mini.cooperst",),
    "vehicle.ford.mustang": ("vehicle.mustang.mustang",),
}

## Vehicle blueprints

#: blueprints for cars
carModels = [
    "vehicle.audi.a2",
    "vehicle.audi.etron",
    "vehicle.audi.tt",
    "vehicle.bmw.grandtourer",
    "vehicle.chevrolet.impala",
    "vehicle.citroen.c3",
    "vehicle.dodge.charger_police",
    "vehicle.jeep.wrangler_rubicon",
    "vehicle.lincoln.mkz_2017",
    "vehicle.mercedes.coupe",
    "vehicle.mini.cooper_s",
    "vehicle.ford.mustang",
    "vehicle.nissan.micra",
    "vehicle.nissan.patrol",
    "vehicle.seat.leon",
    "vehicle.tesla.model3",
    "vehicle.toyota.prius",
    "vehicle.volkswagen.t2",
]

#: blueprints for bicycles
bicycleModels = [
    "vehicle.bh.crossbike",
    "vehicle.diamondback.century",
    "vehicle.gazelle.omafiets",
]

#: blueprints for motorcycles
motorcycleModels = [
    "vehicle.harley-davidson.low_rider",
    "vehicle.kawasaki.ninja",
    "vehicle.yamaha.yzf",
]

#: blueprints for trucks
truckModels = [
    "vehicle.carlamotors.carlacola",
    "vehicle.tesla.cybertruck",
]

## Prop blueprints

#: blueprints for trash cans
trashModels = [
    "static.prop.trashcan01",
    "static.prop.trashcan02",
    "static.prop.trashcan03",
    "static.prop.trashcan04",
    "static.prop.trashcan05",
    "static.prop.bin",
]

#: blueprints for traffic cones
coneModels = [
    "static.prop.constructioncone",
    "static.prop.trafficcone01",
    "static.prop.trafficcone02",
]

#: blueprints for road debris
debrisModels = [
    "static.prop.dirtdebris01",
    "static.prop.dirtdebris02",
    "static.prop.dirtdebris03",
]

#: blueprints for vending machines
vendingMachineModels = [
    "static.prop.vendingmachine",
]

#: blueprints for chairs
chairModels = [
    "static.prop.plasticchair",
]

#: blueprints for bus stops
busStopModels = [
    "static.prop.busstop",
]

#: blueprints for roadside billboards
advertisementModels = [
    "static.prop.advertisement",
    "static.prop.streetsign",
    "static.prop.streetsign01",
    "static.prop.streetsign04",
]

#: blueprints for pieces of trash
garbageModels = [
    "static.prop.colacan",
    "static.prop.garbage01",
    "static.prop.garbage02",
    "static.prop.garbage03",
    "static.prop.garbage04",
    "static.prop.garbage05",
    "static.prop.garbage06",
    "static.prop.plasticbag",
    "static.prop.trashbag",
]

#: blueprints for containers
containerModels = [
    "static.prop.container",
    "static.prop.clothcontainer",
    "static.prop.glasscontainer",
]

#: blueprints for tables
tableModels = [
    "static.prop.table",
    "static.prop.plastictable",
]

#: blueprints for traffic barriers
barrierModels = [
    "static.prop.streetbarrier",
    "static.prop.chainbarrier",
    "static.prop.chainbarrierend",
]

#: blueprints for flowerpots
plantpotModels = [
    "static.prop.plantpot01",
    "static.prop.plantpot02",
    "static.prop.plantpot03",
    "static.prop.plantpot04",
    "static.prop.plantpot05",
    "static.prop.plantpot06",
    "static.prop.plantpot07",
    "static.prop.plantpot08",
]

#: blueprints for mailboxes
mailboxModels = [
    "static.prop.mailbox",
]

#: blueprints for garden gnomes
gnomeModels = [
    "static.prop.gnome",
]

#: blueprints for creased boxes
creasedboxModels = [
    "static.prop.creasedbox01",
    "static.prop.creasedbox02",
    "static.prop.creasedbox03",
]

#: blueprints for briefcases, suitcases, etc.
caseModels = [
    "static.prop.travelcase",
    "static.prop.briefcase",
    "static.prop.guitarcase",
]

#: blueprints for boxes
boxModels = [
    "static.prop.box01",
    "static.prop.box02",
    "static.prop.box03",
]

#: blueprints for benches
benchModels = [
    "static.prop.bench01",
    "static.prop.bench02",
    "static.prop.bench03",
]

#: blueprints for barrels
barrelModels = [
    "static.prop.barrel",
]

#: blueprints for ATMs
atmModels = [
    "static.prop.atm",
]

#: blueprints for kiosks
kioskModels = [
    "static.prop.kiosk_01",
]

#: blueprints for iron plates
ironplateModels = [
    "static.prop.ironplank",
]

#: blueprints for traffic warning signs
trafficwarningModels = [
    "static.prop.trafficwarning",
]

## Walker blueprints

#: blueprints for pedestrians
walkerModels = [
    "walker.pedestrian.0001",
    "walker.pedestrian.0002",
    "walker.pedestrian.0003",
    "walker.pedestrian.0004",
    "walker.pedestrian.0005",
    "walker.pedestrian.0006",
    "walker.pedestrian.0007",
    "walker.pedestrian.0008",
    "walker.pedestrian.0009",
    "walker.pedestrian.0010",
    "walker.pedestrian.0011",
    "walker.pedestrian.0012",
    "walker.pedestrian.0013",
    "walker.pedestrian.0014",
]
